import collections
import enum
import math
import pathlib
import types
from typing import Any, Literal, Mapping, Optional, Sequence, Tuple, Union

from scvelo.plotting import paga

import numpy as np
import pandas as pd
import scipy.sparse as sp

import matplotlib as mpl
import matplotlib.pyplot as plt
import seaborn as sns
from matplotlib import cm, colors

from anndata import AnnData
from scanpy.plotting import violin

from cellrank import logging as logg
from cellrank._utils import Lineage
from cellrank._utils._docs import d, inject_docs
from cellrank._utils._enum import ModeEnum
from cellrank._utils._key import Key
from cellrank._utils._utils import (
    RandomKeys,
    _unique_order_preserving,
    save_fig,
    valuedispatch,
)
from cellrank.pl._utils import _position_legend

__all__ = ["aggregate_fate_probabilities"]


class AggregationMode(ModeEnum):
    BAR = enum.auto()
    PAGA = enum.auto()
    PAGA_PIE = enum.auto()
    VIOLIN = enum.auto()
    HEATMAP = enum.auto()
    CLUSTERMAP = enum.auto()


@d.dedent
@inject_docs(m=AggregationMode)
def aggregate_fate_probabilities(
    adata: AnnData,
    mode: Literal["bar", "paga", "paga_pie", "violin", "heatmap", "clustermap"] = AggregationMode.PAGA_PIE,
    backward: bool = False,
    lineages: Optional[Union[str, Sequence[str]]] = None,
    cluster_key: Optional[str] = "clusters",
    clusters: Optional[Union[str, Sequence[str]]] = None,
    basis: Optional[str] = None,
    cbar: bool = True,
    ncols: Optional[int] = None,
    sharey: bool = False,
    fmt: str = "0.2f",
    xrot: float = 90,
    legend_kwargs: Mapping[str, Any] = types.MappingProxyType({"loc": "best"}),
    figsize: Optional[Tuple[float, float]] = None,
    dpi: Optional[int] = None,
    save: Optional[Union[str, pathlib.Path]] = None,
    **kwargs: Any,
) -> None:
    """Plot aggregate lineage probabilities at a cluster level.

    This can be used to investigate how likely a certain cluster is to go to the %(terminal)s states,
    or in turn to have descended from the %(initial)s states.
    For mode `{m.PAGA!r}` and `{m.PAGA_PIE!r}`, we use *PAGA* :cite:`wolf:19`.

    Parameters
    ----------
    %(adata)s
    mode
        Type of plot to show. Valid options are:

        - ``{m.BAR!r}`` - barplot, one panel per cluster. The whiskers correspond to the standard error of the mean.
        - ``{m.PAGA!r}`` - :func:`~scvelo.pl.paga`, one per %(initial_or_terminal)s state, colored in by fate.
        - ``{m.PAGA_PIE!r}`` - :func:`~scvelo.pl.paga` with pie charts indicating aggregated fates.
        - ``{m.VIOLIN!r}`` - violin plots, one per %(initial_or_terminal)s state.
        - ``{m.HEATMAP!r}`` - a heatmap, showing average fates per cluster.
        - ``{m.CLUSTERMAP!r}`` - same as a heatmap, but with a dendrogram.
    %(backward)s
    lineages
        Lineages for which to visualize the fate probabilities. If :obj:`None`, use all lineages.
    cluster_key
        Key in :attr:`~anndata.AnnData.obs` containing the clusters.
    clusters
        Clusters to visualize. If :obj:`None`, all clusters will be plotted.
    basis
        Basis for scatterplot to use when ``mode = {m.PAGA_PIE!r}``. If :obj:`None`, don't show the scatterplot.
    cbar
        Whether to show colorbar when ``mode = {m.PAGA_PIE!r}``.
    ncols
        Number of columns when ``mode = {m.BAR!r}`` or ``mode = {m.PAGA!r}``.
    sharey
        Whether to share y-axis when ``mode = {m.BAR!r}``.
    fmt
        Format when using ``mode = {m.HEATMAP!r}`` or ``mode = {m.CLUSTERMAP!r}``.
    xrot
        Rotation of the labels on the x-axis.
    figsize
        Size of the figure.
    legend_kwargs
        Keyword arguments for :meth:`~matplotlib.axes.Axes.legend`.
        For ``mode = {m.PAGA_PIE!r}`` and ``basis = '...'``, this controls the placement of the
        fate probabilities legend.
    %(plotting)s
    kwargs
        Keyword arguments for :func:`~scvelo.pl.paga`, :func:`~scanpy.pl.violin` or
        :func:`~matplotlib.pyplot.bar`, depending on the ``mode``.

    Returns
    -------
    %(just_plots)s
    """

    @valuedispatch
    def plot(mode: AggregationMode, *_args, **_kwargs):
        raise NotImplementedError(mode.value)

    @plot.register(AggregationMode.BAR)
    def _():
        cols = 4 if ncols is None else ncols
        n_rows = math.ceil(len(clusters) / cols)
        fig = plt.figure(None, (3.5 * cols, 5 * n_rows) if figsize is None else figsize, dpi=dpi)
        fig.tight_layout()

        gs = plt.GridSpec(n_rows, cols, figure=fig, wspace=0.5, hspace=0.5)

        ax = None
        colors = list(probs.colors)

        for g, k in zip(gs, d.keys()):
            current_ax = fig.add_subplot(g, sharey=ax)
            current_ax.bar(
                x=np.arange(probs.nlin),
                height=d[k][0],
                color=colors,
                yerr=d[k][1],
                ecolor="black",
                capsize=10,
                **kwargs,
            )
            if sharey:
                ax = current_ax

            current_ax.set_xticks(np.arange(probs.nlin))
            current_ax.set_xticklabels(probs.names, rotation=xrot)
            if not is_all:
                current_ax.set_xlabel(term_states)
            current_ax.set_ylabel("fate probability")
            current_ax.set_title(k)

        return fig

    @plot.register(AggregationMode.PAGA)
    def _():
        kwargs["save"] = None
        kwargs["show"] = False
        if "cmap" not in kwargs:
            kwargs["cmap"] = cm.viridis

        cols = probs.nlin if ncols is None else ncols
        nrows = math.ceil(probs.nlin / cols)
        fig, axes = plt.subplots(
            nrows,
            cols,
            figsize=(7 * cols, 4 * nrows) if figsize is None else figsize,
            constrained_layout=True,
            dpi=dpi,
        )
        # fig.tight_layout()  can't use this because colorbar.make_axes fails

        i = 0
        axes = [axes] if not isinstance(axes, np.ndarray) else np.ravel(axes)
        vmin, vmax = np.inf, -np.inf

        if basis is not None:
            kwargs["basis"] = basis
            kwargs["scatter_flag"] = True
            kwargs["color"] = cluster_key

        for i, (ax, lineage_name) in enumerate(zip(axes, probs.names)):
            cols = [v[0][i] for v in d.values()]
            kwargs["ax"] = ax
            kwargs["colors"] = tuple(cols)
            kwargs["title"] = f"{direction} {lineage_name}"

            vmin = np.min(cols + [vmin])
            vmax = np.max(cols + [vmax])

            paga(adata, **kwargs)

        if cbar:
            norm = colors.Normalize(vmin=vmin, vmax=vmax)
            cax, _ = mpl.colorbar.make_axes(ax, aspect=60)
            _ = mpl.colorbar.ColorbarBase(
                cax,
                ticks=np.linspace(norm.vmin, norm.vmax, 5),
                norm=norm,
                cmap=kwargs["cmap"],
                label="average fate probability",
            )

        for ax in axes[i + 1 :]:
            ax.remove()

        return fig

    @plot.register(AggregationMode.PAGA_PIE)
    def _():
        colors = {i: collections.OrderedDict(zip(probs.colors, mean)) for i, (mean, _) in enumerate(d.values())}

        fig, ax = plt.subplots(figsize=figsize, dpi=dpi)
        fig.tight_layout()

        kwargs["ax"] = ax
        kwargs["show"] = False
        kwargs["colorbar"] = False  # has to be disabled
        kwargs["show"] = False

        kwargs["node_colors"] = colors
        kwargs.pop("save", None)  # we will handle saving

        kwargs["transitions"] = kwargs.get("transitions", "transitions_confidence")
        if "legend_loc" in kwargs:
            orig_ll = kwargs["legend_loc"]
            if orig_ll != "on data":
                kwargs["legend_loc"] = "none"  # we will handle legend
        else:
            orig_ll = None
            kwargs["legend_loc"] = "on data"

        if basis is not None:
            kwargs["basis"] = basis
            kwargs["scatter_flag"] = True
            kwargs["color"] = cluster_key

        ax = paga(adata, **kwargs)
        ax.set_title(kwargs.get("title", cluster_key))

        if basis is not None and orig_ll not in ("none", "on data", None):
            handles = []
            for cluster_name, color in zip(
                adata.obs[f"{cluster_key}"].cat.categories,
                adata.uns[f"{cluster_key}_colors"],
            ):
                handles += [ax.scatter([], [], label=cluster_name, c=color)]
            first_legend = _position_legend(
                ax,
                legend_loc=orig_ll,
                handles=handles,
                **{k: v for k, v in legend_kwargs.items() if k != "loc"},
                title=cluster_key,
            )
            fig.add_artist(first_legend)

        if legend_kwargs.get("loc", None) not in ("none", "on data", None):
            # we need to use these, because scvelo can have its own handles and
            # they would be plotted here
            handles = []
            for lineage_name, color in zip(probs.names, colors[0].keys()):
                handles += [ax.scatter([], [], label=lineage_name, c=color)]
            if len(colors[0].keys()) != Lineage.from_adata(adata, backward=backward).nlin:
                handles += [ax.scatter([], [], label="Rest", c="grey")]

            second_legend = _position_legend(
                ax,
                legend_loc=legend_kwargs["loc"],
                handles=handles,
                **{k: v for k, v in legend_kwargs.items() if k != "loc"},
                title=term_states,
            )
            fig.add_artist(second_legend)

        return fig

    @plot.register(AggregationMode.VIOLIN)
    def _():
        kwargs.pop("ax", None)
        kwargs.pop("keys", None)
        kwargs.pop("save", None)  # we will handle saving

        kwargs["show"] = False
        kwargs["groupby"] = cluster_key
        kwargs["rotation"] = xrot

        cols = probs.nlin if ncols is None else ncols
        nrows = math.ceil(probs.nlin / cols)

        fig, axes = plt.subplots(
            nrows,
            cols,
            figsize=(6 * cols, 4 * nrows) if figsize is None else figsize,
            sharey=sharey,
            dpi=dpi,
        )
        fig.tight_layout()
        fig.subplots_adjust(wspace=0.2, hspace=0.3)

        if not isinstance(axes, np.ndarray):
            axes = [axes]
        axes = np.ravel(axes)

        with RandomKeys(adata, probs.nlin, where="obs") as keys:
            _i = 0
            for _i, (name, key, ax) in enumerate(zip(probs.names, keys, axes)):
                adata.obs[key] = probs[name].X[:, 0]
                ax.set_title(f"{direction} {name}")
                violin(adata, ylabel="fate probability", keys=key, ax=ax, **kwargs)
            for ax in axes[_i + 1 :]:
                ax.remove()

        return fig

    def plot_violin_no_cluster_key():
        kwargs.pop("ax", None)
        kwargs.pop("keys", None)  # don't care
        kwargs.pop("save", None)

        kwargs["show"] = False
        kwargs["groupby"] = term_states
        kwargs["xlabel"] = None
        kwargs["rotation"] = xrot

        data = np.ravel(probs.X.T)[..., None]
        tmp = AnnData(sp.csr_matrix(data.shape, dtype=data.dtype))
        tmp.obs["fate probability"] = data
        tmp.obs[term_states] = (
            pd.Series(np.concatenate([[f"{direction.lower()} {n}"] * adata.n_obs for n in probs.names]))
            .astype("category")
            .values
        )
        tmp.obs[term_states] = tmp.obs[term_states].cat.reorder_categories(
            [f"{direction.lower()} {n}" for n in probs.names]
        )
        tmp.uns[f"{term_states}_colors"] = probs.colors

        fig, ax = plt.subplots(figsize=figsize if figsize is not None else (8, 6), dpi=dpi)
        ax.set_title(term_states.capitalize())

        violin(tmp, keys=["fate probability"], ax=ax, **kwargs)

        return fig

    @plot.register(AggregationMode.HEATMAP)
    def _():
        data = pd.DataFrame([mean for mean, _ in d.values()], columns=probs.names, index=clusters).T

        title = kwargs.pop("title", "average fate per cluster")
        vmin, vmax = data.values.min(), data.values.max()
        cbar_kws = {
            "label": "probability",
            "ticks": np.linspace(vmin, vmax, 5),
            "format": "%.3f",
        }
        kwargs.setdefault("cmap", "viridis")

        if use_clustermap:
            kwargs["cbar_pos"] = (0, 0.9, 0.025, 0.15) if cbar else None
            max_size = float(max(data.shape))

            g = sns.clustermap(
                data=data,
                annot=True,
                vmin=vmin,
                vmax=vmax,
                fmt=fmt,
                row_colors=probs.colors,
                dendrogram_ratio=(
                    0.15 * data.shape[0] / max_size,
                    0.15 * data.shape[1] / max_size,
                ),
                cbar_kws=cbar_kws,
                figsize=figsize,
                **kwargs,
            )
            g.ax_heatmap.set_xlabel(cluster_key)
            g.ax_heatmap.set_ylabel("lineage")
            g.ax_col_dendrogram.set_title(title)

            fig = g.fig
            g = g.ax_heatmap
        else:
            fig, ax = plt.subplots(figsize=figsize, dpi=dpi)
            g = sns.heatmap(
                data=data,
                vmin=vmin,
                vmax=vmax,
                annot=True,
                fmt=fmt,
                cbar=cbar,
                cbar_kws=cbar_kws,
                ax=ax,
                **kwargs,
            )
            ax.set_title(title)
            ax.set_xlabel(cluster_key)
            ax.set_ylabel("lineage")

        g.set_xticklabels(g.get_xticklabels(), rotation=xrot)
        g.set_yticklabels(g.get_yticklabels(), rotation=0)

        return fig

    mode = AggregationMode(mode)

    if cluster_key is not None:
        if cluster_key not in adata.obs:
            raise KeyError(f"Key `{cluster_key!r}` not found in `adata.obs`.")
    elif mode not in (mode.BAR, mode.VIOLIN):
        raise ValueError(
            f"Not specifying cluster key is only available for modes "
            f"`{AggregationMode.BAR!r}` and `{AggregationMode.VIOLIN!r}`, found `mode={mode!r}`."
        )

    term_states = Key.obs.term_states(backward)
    direction = Key.where(backward)
    if cluster_key is not None:
        is_all = False
        if clusters is not None:
            if isinstance(clusters, str):
                clusters = [clusters]
            clusters = _unique_order_preserving(clusters)
            if mode in (mode.PAGA, mode.PAGA_PIE):
                logg.debug(f"Setting `clusters` to all available ones because of `mode={mode!r}`")
                clusters = list(adata.obs[cluster_key].cat.categories)
            else:
                for cname in clusters:
                    if cname not in adata.obs[cluster_key].cat.categories:
                        raise KeyError(f"Cluster `{cname!r}` not found in `adata.obs[{cluster_key!r}]`.")
        else:
            clusters = list(adata.obs[cluster_key].cat.categories)
    else:
        is_all = True
        clusters = [term_states]

    probs = Lineage.from_adata(adata, backward=backward)
    if lineages is None:
        # must be list for `sc.pl.violin`, else cats str
        lineages = list(probs.names)
    elif isinstance(lineages, str):
        lineages = [lineages]
    lineages = _unique_order_preserving(lineages)
    probs = probs[:, lineages]

    if mode == mode.VIOLIN and not is_all:
        mask = np.isin(adata.obs[cluster_key], clusters)
        adata = adata[mask].copy()
        probs = probs[mask]

    d = collections.OrderedDict()
    for name in clusters:
        data = probs.X if is_all else probs.X[(adata.obs[cluster_key] == name).to_numpy()]
        mean = np.nanmean(data, axis=0)
        std = np.nanstd(data, axis=0) / np.sqrt(data.shape[0])
        d[name] = [mean, std]

    logg.debug(f"Plotting in mode `{mode!r}`")
    use_clustermap = False
    if mode == mode.CLUSTERMAP:
        use_clustermap = True
        mode = mode.HEATMAP
    elif mode in (AggregationMode.PAGA, AggregationMode.PAGA_PIE) and "paga" not in adata.uns:
        raise KeyError("Compute PAGA first as `scvelo.tl.paga()` or `scanpy.tl.paga()`.")

    fig = plot_violin_no_cluster_key() if mode == AggregationMode.VIOLIN and cluster_key is None else plot(mode)

    if save is not None:
        save_fig(fig, save)



import abc
from typing import Any, Callable, Optional, Tuple

import numpy as np
import scipy.sparse as sp

from cellrank._utils._docs import d
from cellrank._utils._parallelize import parallelize

__all__ = ["HardThresholdScheme", "SoftThresholdScheme", "CustomThresholdScheme"]


class ThresholdSchemeABC(abc.ABC):
    """Base class for all connectivity biasing schemes."""

    @d.get_summary(base="pt_scheme")
    @d.get_sections(base="pt_scheme", sections=["Parameters", "Returns"])
    @abc.abstractmethod
    def __call__(
        self,
        cell_pseudotime: float,
        neigh_pseudotime: np.ndarray,
        neigh_conn: np.ndarray,
        **kwargs: Any,
    ) -> np.ndarray:
        """Calculate biased connections for a given cell.

        Parameters
        ----------
        cell_pseudotime
            Pseudotime of the current cell.
        neigh_pseudotime
            Array of shape ``(n_neighbors,)`` containing pseudotime of neighbors.
        neigh_conn
            Array of shape ``(n_neighbors,)`` containing connectivities of the current cell and its neighbors.

        Returns
        -------
        Array of shape ``(n_neighbors,)`` containing the biased connectivities.
        """

    def _bias_knn_helper(
        self,
        ixs: np.ndarray,
        conn: sp.csr_matrix,
        pseudotime: np.ndarray,
        queue=None,
        **kwargs: Any,
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        i, indices, indptr, data = 0, [], [], []
        for i in ixs:
            row = conn[i]
            biased_row = self(pseudotime[i], pseudotime[row.indices], row.data, **kwargs)
            if np.shape(biased_row) != row.data.shape:
                raise ValueError(f"Expected row of shape `{row.data.shape}`, found `{np.shape(biased_row)}`.")

            data.extend(biased_row)
            indices.extend(row.indices)
            indptr.append(conn.indptr[i])

            if queue is not None:
                queue.put(1)

        if i == conn.shape[0] - 1:
            indptr.append(conn.indptr[-1])
        if queue is not None:
            queue.put(None)

        return np.array(data), np.array(indices), np.array(indptr)

    @d.dedent
    def bias_knn(
        self,
        conn: sp.csr_matrix,
        pseudotime: np.ndarray,
        n_jobs: Optional[int] = None,
        backend: str = "loky",
        show_progress_bar: bool = True,
        **kwargs: Any,
    ) -> sp.csr_matrix:
        """Bias cell-cell connectivities of a KNN graph.

        Parameters
        ----------
        conn
            Sparse matrix of shape ``(n_cells, n_cells)`` containing the nearest neighbor connectivities.
        pseudotime
            Pseudotemporal ordering of cells.
        %(parallel)s

        Returns
        -------
        The biased connectivities.
        """
        res = parallelize(
            self._bias_knn_helper,
            np.arange(conn.shape[0]),
            as_array=False,
            unit="cell",
            n_jobs=n_jobs,
            backend=backend,
            show_progress_bar=show_progress_bar,
        )(conn, pseudotime, **kwargs)
        data, indices, indptr = zip(*res)

        conn = sp.csr_matrix((np.concatenate(data), np.concatenate(indices), np.concatenate(indptr)))
        conn.eliminate_zeros()

        return conn

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}>"

    def __str__(self) -> str:
        return repr(self)


class HardThresholdScheme(ThresholdSchemeABC):
    """Thresholding scheme inspired by *Palantir* :cite:`setty:19`.

    Note that this won't exactly reproduce the original *Palantir* results:

    - *Palantir* computes the kNN graph in a scaled space of diffusion components.
    - *Palantir* uses its own pseudotime to bias the kNN graph which is not implemented here.
    - *Palantir* uses a slightly different mechanism to ensure the graph remains connected when removing edges
      that point into the "pseudotime past".
    """

    @d.dedent
    def __call__(
        self,
        cell_pseudotime: float,
        neigh_pseudotime: np.ndarray,
        neigh_conn: np.ndarray,
        frac_to_keep: float = 0.3,
    ) -> np.ndarray:
        """Convert the undirected graph of cell-cell similarities into a directed one by removing "past" edges.

        This uses a pseudotemporal measure to remove graph-edges that point into the pseudotime-past. For each cell,
        it keeps the closest neighbors, even if they are in the pseudotime past, to make sure the graph remains
        connected.

        Parameters
        ----------
        %(pt_scheme.parameters)s
        frac_to_keep
            The ``frac_to_keep`` * n_neighbors closest neighbors (according to graph connectivities) are kept, no matter
            whether they lie in the pseudotemporal past or future. Must be in :math:`[0, 1]`.

        Returns
        -------
        %(pt_scheme.returns)s
        """
        if not (0 <= frac_to_keep <= 1):
            raise ValueError(f"Expected `frac_to_keep` to be in `[0, 1]`, found `{frac_to_keep}`.")

        k_thresh = max(0, min(30, int(np.floor(len(neigh_conn) * frac_to_keep))))
        ixs = np.flip(np.argsort(neigh_conn))
        close_ixs, far_ixs = ixs[:k_thresh], ixs[k_thresh:]

        mask_keep = cell_pseudotime <= neigh_pseudotime[far_ixs]
        far_ixs_keep = far_ixs[mask_keep]

        biased_conn = np.zeros_like(neigh_conn)
        biased_conn[close_ixs] = neigh_conn[close_ixs]
        biased_conn[far_ixs_keep] = neigh_conn[far_ixs_keep]

        return biased_conn


class SoftThresholdScheme(ThresholdSchemeABC):
    """Thresholding scheme inspired by :cite:`stassen:21`.

    The idea is to downweight edges that points against the direction of increasing pseudotime. Essentially, the
    further "behind" a query cell is in pseudotime with respect to the current reference cell, the more penalized will
    be its graph-connectivity.
    """

    @d.dedent
    def __call__(
        self,
        cell_pseudotime: float,
        neigh_pseudotime: np.ndarray,
        neigh_conn: np.ndarray,
        b: float = 10.0,
        nu: float = 0.5,
    ) -> np.ndarray:
        """Bias the connectivities by downweighting ones to past cells.

        This function uses `generalized logistic regression
        <https://en.wikipedia.org/wiki/Generalized_logistic_function>`_ to weight the past connectivities.

        Parameters
        ----------
        %(pt_scheme.parameters)s
        %(soft_scheme)s

        Returns
        -------
        %(pt_scheme.returns)s
        """
        past_ixs = np.where(neigh_pseudotime < cell_pseudotime)[0]
        if not len(past_ixs):
            return neigh_conn

        weights = np.ones_like(neigh_conn)

        dt = cell_pseudotime - neigh_pseudotime[past_ixs]
        weights[past_ixs] = 2.0 / ((1.0 + np.exp(b * dt)) ** (1.0 / nu))

        return neigh_conn * weights


class CustomThresholdScheme(ThresholdSchemeABC):
    """Class that wraps a user supplied scheme.

    Parameters
    ----------
    callback
        Function which returns the biased connectivities.
    """

    def __init__(
        self,
        callback: Callable[[float, np.ndarray, np.ndarray, np.ndarray, Any], np.ndarray],
    ):
        super().__init__()
        self._callback = callback

    @d.dedent
    def __call__(
        self,
        cell_pseudotime: float,
        neigh_pseudotime: np.ndarray,
        neigh_conn: np.ndarray,
        **kwargs: Any,
    ) -> np.ndarray:
        """%(pt_scheme.summary)s

        Parameters
        ----------
        %(pt_scheme.parameters)s
        kwargs
            Additional keyword arguments.

        Returns
        -------
        %(pt_scheme.returns)s
        """  # noqa: D400
        return self._callback(cell_pseudotime, neigh_pseudotime, neigh_conn, **kwargs)



import pathlib
import types
from typing import Any, List, Mapping, Optional, Sequence, Tuple, Union

import numpy as np
import pandas as pd

import matplotlib
import matplotlib.pyplot as plt
from matplotlib import cm

from anndata import AnnData

from cellrank import logging as logg
from cellrank._utils import Lineage
from cellrank._utils._docs import d
from cellrank._utils._enum import DEFAULT_BACKEND, Backend_t
from cellrank._utils._parallelize import _get_n_cores
from cellrank._utils._utils import (
    _check_collection,
    _genesymbols,
    _unique_order_preserving,
    save_fig,
)
from cellrank.pl._utils import (
    _callback_type,
    _create_callbacks,
    _create_models,
    _fit_bulk,
    _get_backend,
    _input_model_type,
    _return_model_type,
    _time_range_type,
    _trends_helper,
)

__all__ = ["gene_trends"]


@d.dedent
@_genesymbols
def gene_trends(
    adata: AnnData,
    model: _input_model_type,
    genes: Union[str, Sequence[str]],
    time_key: str,
    lineages: Optional[Union[str, Sequence[str]]] = None,
    backward: bool = False,
    data_key: str = "X",
    time_range: Optional[Union[_time_range_type, List[_time_range_type]]] = None,
    transpose: bool = False,
    callback: _callback_type = None,
    conf_int: Union[bool, float] = True,
    same_plot: bool = False,
    hide_cells: bool = False,
    perc: Optional[Union[Tuple[float, float], Sequence[Tuple[float, float]]]] = None,
    lineage_cmap: Optional[matplotlib.colors.ListedColormap] = None,
    fate_prob_cmap: matplotlib.colors.ListedColormap = cm.viridis,
    cell_color: Optional[str] = None,
    cell_alpha: float = 0.6,
    lineage_alpha: float = 0.2,
    size: float = 15,
    lw: float = 2,
    cbar: bool = True,
    margins: float = 0.015,
    sharex: Optional[Union[str, bool]] = None,
    sharey: Optional[Union[str, bool]] = None,
    gene_as_title: Optional[bool] = None,
    legend_loc: Optional[str] = "best",
    obs_legend_loc: Optional[str] = "best",
    ncols: int = 2,
    suptitle: Optional[str] = None,
    return_models: bool = False,
    n_jobs: Optional[int] = 1,
    backend: Backend_t = DEFAULT_BACKEND,
    show_progress_bar: bool = True,
    figsize: Optional[Tuple[float, float]] = None,
    dpi: Optional[int] = None,
    save: Optional[Union[str, pathlib.Path]] = None,
    return_figure: bool = False,
    plot_kwargs: Mapping[str, Any] = types.MappingProxyType({}),
    **kwargs: Any,
) -> Optional[_return_model_type]:
    """Plot gene expression trends along lineages.

    .. seealso::
        - See :doc:`../../../notebooks/tutorials/estimators/800_gene_trends` on how to
          visualize the gene trends.

    Each lineage is defined via its lineage weights. This function accepts any model based off
    :class:`~cellrank.models.BaseModel` to fit gene expression, where we take the lineage weights
    into account in the loss function.

    Parameters
    ----------
    %(adata)s
    %(model)s
    %(genes)s
    time_key
        Key in :attr:`~anndata.AnnData.obs` where the pseudotime is stored.
    lineages
        Names of the lineages to plot. If :obj:`None`, plot all lineages.
    %(backward)s
    data_key
        Key in :attr:`~anndata.AnnData.layers` or ``'X'`` for :attr:`~anndata.AnnData.X` where the data is stored.
    %(time_range)s
        This can also be specified on per-lineage basis.
    %(gene_symbols)s
    transpose
        If ``same_plot = True``, group the trends by ``lineages`` instead of ``genes``.
        This forces ``hide_cells = True``.
        If ``same_plot = False``, show ``lineages`` in rows and ``genes`` in columns.
    %(model_callback)s
    conf_int
        Whether to compute and show confidence interval. If the ``model`` is :class:`~cellrank.models.GAMR`,
        it can also specify the confidence level, the default is :math:`0.95`.
    same_plot
        Whether to plot all lineages for each gene in the same plot.
    hide_cells
        If :obj:`True`, hide all cells.
    perc
        Percentile for colors. Valid values are in :math:`[0, 100]`.
        This can improve visualization. Can be specified individually for each lineage.
    lineage_cmap
        Categorical colormap to use when coloring in the lineages. If :obj:`None` and ``same_plot = True``,
        use the corresponding colors in :attr:`~anndata.AnnData.uns`, otherwise use ``'black'``.
    fate_prob_cmap
        Continuous colormap to use when visualizing the fate probabilities for each lineage.
        Only used when ``same_plot = False``.
    cell_color
        Key in :attr:`~anndata.AnnData.obs` or :attr:`~anndata.AnnData.var_names` used for coloring the cells.
    cell_alpha
        Alpha channel for cells.
    lineage_alpha
        Alpha channel for lineage confidence intervals.
    size
        Size of the points.
    lw
        Line width of the smoothed values.
    cbar
        Whether to show colorbar. Always shown when percentiles for lineages differ.
        Only used when ``same_plot = False``.
    margins
        Margins around the plot.
    sharex
        Whether to share x-axis. Valid options are ``'row'``, ``'col'`` or ``'none'``.
    sharey
        Whether to share y-axis. Valid options are ``'row'`, ``'col'`` or ``'none'``.
    gene_as_title
        Whether to show gene names as titles instead on y-axis.
    legend_loc
        Location of the legend displaying lineages. Only used when ``same_plot = True``.
    obs_legend_loc
        Location of the legend when ``cell_color`` corresponds to a categorical variable.
    ncols
        Number of columns of the plot when plotting multiple genes. Only used when ``same_plot = True``.
    suptitle
        Suptitle of the figure.
    return_figure
        Whether to return the figure object.
    %(return_models)s
    %(parallel)s
    %(plotting)s
    plot_kwargs
        Keyword arguments for the :meth:`~cellrank.models.BaseModel.plot`.
    kwargs
        Keyword arguments for :meth:`~cellrank.models.BaseModel.prepare`.

    Returns
    -------
    %(plots_or_returns_models)s
    """
    if isinstance(genes, str):
        genes = [genes]
    genes = _unique_order_preserving(genes)

    _check_collection(
        adata,
        genes,
        "obs" if data_key == "obs" else "var_names",
        use_raw=kwargs.get("use_raw", False),
    )

    probs = Lineage.from_adata(adata, backward=backward)
    if lineages is None:
        lineages = probs.names
    elif isinstance(lineages, str):
        lineages = [lineages]
    elif all(ln is None for ln in lineages):  # no lineage, all the weights are 1
        lineages = [None]
        cbar = False
        logg.debug("All lineages are `None`, setting the weights to `1`")
    lineages = _unique_order_preserving(lineages)

    if isinstance(time_range, (tuple, float, int, type(None))):
        time_range = [time_range] * len(lineages)
    elif len(time_range) != len(lineages):
        raise ValueError(f"Expected time ranges to be of length `{len(lineages)}`, found `{len(time_range)}`.")

    kwargs["time_key"] = time_key
    kwargs["data_key"] = data_key
    kwargs["backward"] = backward
    kwargs["conf_int"] = conf_int  # prepare doesnt take or need this
    models = _create_models(model, genes, lineages)

    all_models, models, genes, lineages = _fit_bulk(
        models,
        _create_callbacks(adata, callback, genes, lineages, **kwargs),
        genes,
        lineages,
        time_range,
        return_models=True,
        filter_all_failed=False,
        parallel_kwargs={
            "show_progress_bar": show_progress_bar,
            "n_jobs": _get_n_cores(n_jobs, len(genes)),
            "backend": _get_backend(models, backend),
        },
        **kwargs,
    )

    lineages = sorted(lineages)
    probs = probs[lineages]
    if lineage_cmap is None and not transpose:
        lineage_cmap = probs.colors

    plot_kwargs = dict(plot_kwargs)
    plot_kwargs["obs_legend_loc"] = obs_legend_loc
    if transpose:
        all_models = pd.DataFrame(all_models).T.to_dict()
        models = pd.DataFrame(models).T.to_dict()
        genes, lineages = lineages, genes
        hide_cells = same_plot or hide_cells
    else:
        # information overload otherwise
        plot_kwargs["lineage_probability"] = False
        plot_kwargs["lineage_probability_conf_int"] = False

    tmp = pd.DataFrame(models).T.astype(bool)
    start_rows = np.argmax(tmp.values, axis=0)
    end_rows = tmp.shape[0] - np.argmax(tmp[::-1].values, axis=0) - 1

    if same_plot:
        gene_as_title = True if gene_as_title is None else gene_as_title
        sharex = "all" if sharex is None else sharex
        if sharey is None:
            sharey = "row" if plot_kwargs.get("lineage_probability", False) else "none"
        ncols = len(genes) if ncols >= len(genes) else ncols
        nrows = int(np.ceil(len(genes) / ncols))
    else:
        gene_as_title = False if gene_as_title is None else gene_as_title
        sharex = "col" if sharex is None else sharex
        if sharey is None:
            sharey = "row" if not hide_cells or plot_kwargs.get("lineage_probability", False) else "none"
        nrows = len(genes)
        ncols = len(lineages)

    plot_kwargs = dict(plot_kwargs)
    if plot_kwargs.get("xlabel", None) is None:
        plot_kwargs["xlabel"] = time_key

    fig, axes = plt.subplots(
        nrows=nrows,
        ncols=ncols,
        sharex=sharex,
        sharey=sharey,
        figsize=(6 * ncols, 4 * nrows) if figsize is None else figsize,
        tight_layout=True,
        dpi=dpi,
    )
    axes = np.reshape(axes, (nrows, ncols))

    cnt = 0
    plot_kwargs["obs_legend_loc"] = None if same_plot else obs_legend_loc

    logg.info("Plotting trends")
    for row in range(len(axes)):
        for col in range(len(axes[row])):
            if cnt >= len(genes):
                break
            gene = genes[cnt]
            if same_plot and plot_kwargs.get("lineage_probability", False) and transpose:
                lpc = probs[gene].colors[0]
            else:
                lpc = None

            if same_plot:
                plot_kwargs["obs_legend_loc"] = obs_legend_loc if row == 0 and col == len(axes[0]) - 1 else None

            _trends_helper(
                models,
                gene=gene,
                lineage_names=lineages,
                transpose=transpose,
                same_plot=same_plot,
                hide_cells=hide_cells,
                perc=perc,
                lineage_cmap=lineage_cmap,
                fate_prob_cmap=fate_prob_cmap,
                lineage_probability_color=lpc,
                cell_color=cell_color,
                alpha=cell_alpha,
                lineage_alpha=lineage_alpha,
                size=size,
                lw=lw,
                cbar=cbar,
                margins=margins,
                sharey=sharey,
                gene_as_title=gene_as_title,
                legend_loc=legend_loc,
                figsize=figsize,
                fig=fig,
                axes=axes[row, col] if same_plot else axes[cnt],
                show_ylabel=col == 0,
                show_lineage=same_plot or (cnt == start_rows),
                show_xticks_and_label=((row + 1) * ncols + col >= len(genes)) if same_plot else (cnt == end_rows),
                **plot_kwargs,
            )
            cnt += 1  # plot legend on the 1st plot

            if not same_plot:
                plot_kwargs["obs_legend_loc"] = None

    if same_plot and (col != ncols):
        for ax in np.ravel(axes)[cnt:]:
            ax.remove()

    fig.suptitle(suptitle, y=1.05)

    if return_figure:
        return fig

    if save is not None:
        save_fig(fig, save)

    if return_models:
        return all_models



from typing import Any, Optional, Tuple, TypeVar, Union

import numpy as np
import scipy.sparse as sp

from anndata import AnnData

__all__ = ["KernelMixin"]

KernelExpression = TypeVar("KernelExpression", bound="KernelMixin")


class KernelMixin:
    """Mixin that exposes various properties of :class:`cellrank.kernels.KernelExpression`."""

    def __init__(self, kernel: "KernelExpression", **kwargs: Any):
        super().__init__(**kwargs)
        self._kernel = kernel
        self._n_obs = self.kernel.adata.n_obs

    @property
    def kernel(self) -> "KernelExpression":
        """Underlying kernel expression."""
        return self._kernel

    @property
    def adata(self) -> AnnData:
        """Annotated data object."""
        return self.kernel.adata

    @property
    def shape(self) -> Tuple[int, int]:
        """Shape of the kernel."""
        return self.kernel.shape

    @adata.setter
    def adata(self, adata: Optional[AnnData]) -> None:
        self.kernel.adata = adata

    def __len__(self) -> int:
        return self._n_obs

    @property
    def transition_matrix(self) -> Union[np.ndarray, sp.spmatrix]:
        """Transition matrix of the :attr:`kernel`."""
        return self.kernel.transition_matrix

    @property
    def backward(self) -> bool:
        """Direction of the :attr:`kernel`."""
        return self.kernel.backward



import itertools
import os
import pathlib
import types
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    Literal,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Union,
)

from tqdm.auto import tqdm

import numpy as np
import pandas as pd
import scipy.sparse as sp
from pandas.api.types import infer_dtype

import scanpy as sc
from anndata import AnnData

from cellrank import logging as logg
from cellrank._utils._docs import d, inject_docs
from cellrank._utils._enum import ModeEnum
from cellrank._utils._utils import _normalize
from cellrank.kernels._base_kernel import UnidirectionalKernel
from cellrank.settings import settings

__all__ = ["RealTimeKernel"]

if TYPE_CHECKING:
    from moscot.problems.spatiotemporal import SpatioTemporalProblem
    from moscot.problems.time import LineageProblem, TemporalProblem


class SelfTransitions(ModeEnum):
    UNIFORM = "uniform"
    DIAGONAL = "diagonal"
    CONNECTIVITIES = "connectivities"
    ALL = "all"


Key_t = Tuple[Any, Any]
Threshold_t = Union[int, float, Literal["auto", "auto_local"]]
Coupling_t = Union[np.ndarray, sp.spmatrix, AnnData]


# TODO(michalk8): subclass the `ExperimentalTimeKernel`
@d.dedent
class RealTimeKernel(UnidirectionalKernel):
    """Kernel which computes transition matrix using optimal transport couplings.

    .. seealso::
        - See :doc:`../../../notebooks/tutorials/kernels/500_real_time` on how to
          compute the :attr:`~cellrank.kernels.RealTimeKernel.transition_matrix` between experimental time points
          using `optimal transport <https://en.wikipedia.org/wiki/Transportation_theory_(mathematics)>`_.

    This class should be constructed using either:

    1. the :meth:`from_moscot` or :meth:`from_wot` method,
    2. explicitly passing the pre-computed couplings, or
    3. overriding the :meth:`compute_coupling` method.

    Parameters
    ----------
    adata
        Annotated data object.
    time_key
        Key in :attr:`~anndata.AnnData.obs` containing the experimental time.
    couplings
        Pre-computed transport couplings. The keys should correspond to a :class:`tuple` of categories from
        the :attr:`time`. If :obj:`None`, the keys will be constructed using the ``policy``
        and the :meth:`compute_coupling` method must be overriden.
    policy
        How to construct the keys from the :attr:`time` when ``couplings = None``:

        - ``policy = 'sequential'`` - the keys will be set to ``[(t1, t2), (t2, t3), ...]``.
        - ``policy = 'triu'`` - the keys will be set to ``[(t1, t2), (t1, t3), ..., (t2, t3), ...]``.
    kwargs
        Keyword arguments for the :class:`~cellrank.kernels.Kernel`.
    """

    def __init__(
        self,
        adata: AnnData,
        time_key: str,
        couplings: Optional[Mapping[Key_t, Optional[Coupling_t]]] = None,
        policy: Literal["sequential", "triu"] = "sequential",
        **kwargs: Any,
    ):
        super().__init__(adata, time_key=time_key, **kwargs)
        self._obs: Optional[pd.DataFrame] = None
        self._couplings, self._reference = self._get_default_coupling(couplings, policy)

    def _read_from_adata(
        self,
        time_key: str,
        **kwargs: Any,
    ) -> None:
        super()._read_from_adata(**kwargs)
        self._time = self.adata.obs[time_key].copy()
        if not isinstance(self._time.dtype, pd.CategoricalDtype):
            raise TypeError(f"Expected `adata.obs[{time_key!r}]` to be categorical, found `{infer_dtype(self._time)}`.")
        self._time = self._time.cat.remove_unused_categories()
        cats = self._time.cat.categories
        self._time_to_ix = dict(zip(cats, range(len(cats))))

    def compute_coupling(self, src: Any, tgt: Any, **kwargs: Any) -> Coupling_t:
        """Compute coupling for a given time pair.

        .. note::
            This implementation only looks up the values in :attr:`couplings`,
            see :meth:`from_moscot` or :meth:`from_wot` on how to initialize them.

        Parameters
        ----------
        src
            Source key in :attr:`time`.
        tgt
            Target key in :attr:`time`.
        kwargs
            Additional keyword arguments.

        Returns
        -------
        Array of shape ``(n_source_cells, n_target_cells)``.
        """
        del kwargs
        if (src, tgt) not in self.couplings:
            raise KeyError(f"Key `{src, tgt}` not found in the `couplings`.")
        if self.couplings[src, tgt] is None:
            raise ValueError(
                f"Coupling for `{src, tgt}` is not computed. " f"Consider overriding the `compute_coupling` method."
            )
        return self.couplings[src, tgt]

    @inject_docs(st=SelfTransitions)
    def compute_transition_matrix(
        self,
        threshold: Optional[Threshold_t] = "auto",
        self_transitions: Union[
            Literal["uniform", "diagonal", "connectivities", "all"],
            Sequence[Any],
        ] = SelfTransitions.CONNECTIVITIES,
        conn_weight: Optional[float] = None,
        conn_kwargs: Mapping[str, Any] = types.MappingProxyType({}),
        **kwargs: Any,
    ) -> "RealTimeKernel":
        """Compute transition matrix from optimal transport couplings.

        Parameters
        ----------
        threshold
            How to remove small non-zero values from the transition matrix. Valid options are:

            - ``'auto'`` - find the maximum threshold value which will not remove every non-zero value from any row.
            - ``'auto_local'`` - same as above, but done for each transport separately.
            - :class:`float` - value in :math:`[0, 100]` corresponding to a percentage of non-zeros to remove in
              the couplings.

            Rows where all values are removed will have a uniform distribution and a warning will be issued.
        self_transitions
            How to define transitions within the blocks that correspond to transitions within the same key.
            Valid options are:

            - ``{st.UNIFORM!r}`` - row-normalized matrix of :math:`1`.
            - ``{st.DIAGONAL!r}`` - identity matrix.
            - ``{st.CONNECTIVITIES!r}`` - transition matrix from the :class:`~cellrank.kernels.ConnectivityKernel`.
            - :class:`~typing.Sequence`` - sequence of source keys defining which blocks should be weighted
              by the connectivities.
            - ``{st.ALL!r}`` - same as above, but for all keys.

            The first 3 options are applied to the block specified by the :attr:`reference`.
        conn_weight
            Weight of connectivities' self transitions. Only used when ``self_transitions = {st.ALL!r}`` or
            a sequence of source keys is passed.
        conn_kwargs
            Keyword arguments for :func:`~scanpy.pp.neighbors` or
            :meth:`~cellrank.kernels.ConnectivityKernel.compute_transition_matrix` when using
            ``self_transitions = 'connectivities'``.
        kwargs
            Keyword arguments for :meth:`compute_coupling`.

        Returns
        -------
        Returns self and updates :attr:`transition_matrix`, :attr:`couplings` and :attr:`params`.
        """
        cache_params = dict(kwargs)
        cache_params["threshold"] = threshold
        cache_params["self_transitions"] = str(self_transitions)
        if self._reuse_cache(cache_params):
            return self

        if (self_transitions == "all" or isinstance(self_transitions, (tuple, list))) and (
            conn_weight is None or not (0 < conn_weight < 1)
        ):
            raise ValueError(f"Expected `conn_weight` to be in interval `(0, 1)`, found `{conn_weight}`.")

        for src, tgt in tqdm(self._couplings, unit="time pair"):
            coupling = self.compute_coupling(src, tgt, **kwargs)
            self.couplings[src, tgt] = self._coupling_to_adata(src, tgt, coupling)

        if threshold is not None:
            self._sparsify_couplings(self.couplings, threshold=threshold, copy=False)

        tmap = self._restich_couplings(
            self.couplings,
            self_transitions=self_transitions,
            conn_weight=conn_weight,
            **conn_kwargs,
        )

        self.transition_matrix = tmap.X
        self._obs = tmap.obs

        return self

    @classmethod
    def from_moscot(
        cls,
        problem: "Union[TemporalProblem, LineageProblem, SpatioTemporalProblem]",
        sparse_mode: Optional[Literal["threshold", "percentile", "min_row"]] = None,
        sparsify_kwargs: Mapping[str, Any] = types.MappingProxyType({}),
        copy: bool = False,
        **kwargs: Any,
    ) -> "RealTimeKernel":
        """Construct the kernel from :mod:`moscot` :cite:`klein:23`.

        Parameters
        ----------
        problem
            :mod:`moscot` problem.
        sparse_mode
            Sparsification mode for :meth:`~moscot.base.output.BaseSolverOutput.sparsify`.
            If :obj:`None`, do not sparsify the outputs. Note that :meth:`compute_transition_matrix`
            can also sparsify the final :attr:`transition_matrix`.
        sparsify_kwargs
            Keyword arguments for the sparsification.
        copy
            Whether to copy the underlying arrays. Note that :class:`jax arrays <jax.Array>` are always copied.
        kwargs
            Keyword arguments for :class:`~cellrank.kernels.RealTimeKernel`.

        Returns
        -------
        The kernel.

        Examples
        --------
        .. code-block:: python

            import moscot as mt
            import cellrank as cr

            adata = mt.datasets.hspc()
            adata.obs["day"] = adata.obs["day"].astype("category")

            problem = mt.problems.TemporalProblem(adata)
            problem = problem.prepare(time_key="day").solve()

            rtk = cr.kernels.RealTimeKernel.from_moscot(problem)
            rtk = rtk.compute_transition_matrix()
        """
        from moscot.utils.subset_policy import SequentialPolicy, TriangularPolicy

        if not problem.solutions:
            raise RuntimeError("Problem contains no solutions. Please run `problem.solve(...)` first.")

        policy = problem._policy
        if isinstance(policy, SequentialPolicy):
            policy = "sequential"
        elif isinstance(policy, TriangularPolicy):
            policy = "triu"
        else:
            raise NotImplementedError(f"Handling `{type(policy)}` policy is not yet implemented.")

        couplings = {}
        for (t1, t2), solution in problem.solutions.items():
            adata_src = problem[t1, t2].adata_src
            adata_tgt = problem[t1, t2].adata_tgt
            if sparse_mode is not None:
                solution = solution.sparsify(mode=sparse_mode, **sparsify_kwargs)
            coupling = solution.transport_matrix
            if not (isinstance(coupling, np.ndarray) or sp.issparse(coupling)):
                # convert from, e.g., `jax`
                # we always copy because otherwise the array is read-only and we may sparsify
                coupling = np.array(coupling, copy=True)
            elif copy:
                coupling = coupling.copy()
            couplings[t1, t2] = AnnData(coupling, obs=adata_src.obs, var=adata_tgt.obs)

        return cls(
            problem.adata,
            couplings=couplings,
            time_key=problem.temporal_key,
            policy=policy,
            **kwargs,
        )

    @classmethod
    def from_wot(
        cls,
        adata: AnnData,
        path: Union[str, pathlib.Path],
        time_key: str,
        **kwargs: Any,
    ) -> "RealTimeKernel":
        """Construct the kernel from Waddington-OT :cite:`schiebinger:19`.

        Parameters
        ----------
        adata
            Annotated data object.
        path
            Directory where the couplings are stored.
        time_key
            Key in :attr:`~anndata.AnnData.obs` containing the experimental time.
        kwargs
            Keyword arguments for :class:`~cellrank.kernels.RealTimeKernel`.

        Returns
        -------
        The kernel.

        Examples
        --------
        .. code-block:: python

            import wot
            import cellrank as cr

            adata = cr.datasets.reprogramming_schiebinger(subset_to_serum=True)
            adata.obs["day"] = adata.obs["day"].astype(float).astype("category")

            ot_model = wot.ot.OTModel(adata, day_field="day")
            ot_model.compute_all_transport_maps(tmap_out="tmaps/")

            rtk = cr.kernels.RealTimeKernel.from_wot(adata, path="tmaps/", time_key="day")
            rtk = rtk.compute_transition_matrix()
        """
        path = pathlib.Path(path)
        dtype = type(adata.obs[time_key].iloc[0])

        couplings = {}
        for fname in path.glob("*h5ad"):
            name, _ = os.path.splitext(fname)
            *_, src, tgt = name.split("_")
            couplings[dtype(src), dtype(tgt)] = sc.read(fname)

        return cls(adata, couplings=couplings, time_key=time_key, policy="sequential", **kwargs)

    @d.dedent
    def _restich_couplings(
        self,
        couplings: Mapping[Key_t, AnnData],
        self_transitions: Union[str, SelfTransitions, Sequence[Any]] = SelfTransitions.DIAGONAL,
        conn_weight: Optional[float] = None,
        **kwargs: Any,
    ) -> AnnData:
        """Group individual transport maps into 1 matrix aligned with :attr:`adata`.

        Parameters
        ----------
        couplings
            Optimal transport couplings.
        self_transitions
            How to define transitions within the blocks that correspond to transitions within the same key.
        conn_weight
            Weight of connectivities' self transitions. Only used when ``self_transitions = {st.ALL!r}`` or
            a sequence of source keys is passed.
        kwargs
            Keyword arguments for :func:`~scanpy.pp.neighbors` or
            :meth:`~cellrank.kernels.ConnectivityKernel.compute_transition_matrix` when using
            ``self_transitions = 'connectivities'``.

        Returns
        -------
        Merged transport maps into one :class:`~anndata.AnnData` object.
        """
        if isinstance(self_transitions, (str, SelfTransitions)):
            self_transitions = SelfTransitions(self_transitions)
            if self_transitions == SelfTransitions.ALL:
                self_transitions = tuple(src for (src, _) in couplings)
        elif isinstance(self_transitions, Iterable):
            self_transitions = tuple(self_transitions)
        else:
            raise TypeError(
                f"Expected `self_transitions` to be a `str` or a `Sequence`, " f"found `{type(self_transitions)}`."
            )
        self._validate_couplings(couplings)

        conn_kwargs = dict(kwargs)
        conn_kwargs["copy"] = False
        _ = conn_kwargs.pop("key_added", None)

        obs_names, obs = {}, {}
        blocks = [[None] * (len(self._time_to_ix)) for _ in range(len(self._time_to_ix))]
        for (src, tgt), coupling in couplings.items():
            src_ix = self._time_to_ix[src]
            tgt_ix = self._time_to_ix[tgt]
            blocks[src_ix][tgt_ix] = coupling.X
            obs_names[src_ix] = coupling.obs_names
            obs[src_ix] = coupling.obs
            # to prevent blocks from disappearing
            n = np.sum(self._time == src)
            blocks[src_ix][src_ix] = sp.spdiags([0] * n, 0, n, n)
            if tgt == self._reference:
                obs_names[tgt_ix] = coupling.var_names

        # if policy='sequential', reference is the last key
        ref_ix = self._time_to_ix[self._reference]
        ref_mask = self._time == self._reference
        n = int(np.sum(ref_mask))

        if self_transitions == SelfTransitions.DIAGONAL:
            blocks[ref_ix][ref_ix] = sp.spdiags([1] * n, 0, n, n)
        elif self_transitions == SelfTransitions.UNIFORM:
            blocks[ref_ix][ref_ix] = np.ones((n, n)) / float(n)
        elif self_transitions == SelfTransitions.CONNECTIVITIES:
            blocks[ref_ix][ref_ix] = _compute_connectivity_tmat(self.adata[ref_mask], **conn_kwargs)
        elif isinstance(self_transitions, tuple):
            verbosity = settings.verbosity
            try:  # ignore overly verbose logging
                settings.verbosity = 0
                for (src, tgt), coupling in couplings.items():
                    # fmt: off
                    if src not in self_transitions:
                        continue
                    src_ix, tgt_ix = self._time_to_ix[src], self._time_to_ix[tgt]
                    blocks[src_ix][src_ix] = conn_weight * _compute_connectivity_tmat(
                        self.adata[coupling.obs_names], **conn_kwargs
                    )
                    blocks[src_ix][tgt_ix] = (1 - conn_weight) * _normalize(blocks[src_ix][tgt_ix])
                    # fmt: on
                blocks[ref_ix][ref_ix] = _compute_connectivity_tmat(self.adata[ref_mask], **conn_kwargs)
            finally:
                settings.verbosity = verbosity
        else:
            raise NotImplementedError(f"Self transitions' mode `{self_transitions}` is not yet implemented.")

        index = []
        for ix in range(len(blocks)):
            index.extend(obs_names[ix])

        tmp = AnnData(sp.bmat(blocks, format="csr"))
        tmp.obs_names = index
        tmp.var_names = index
        tmp = tmp[self.adata.obs_names, :][:, self.adata.obs_names]

        obs = pd.concat([obs.get(ix, None) for ix in range(len(blocks))])
        tmp.obs = pd.merge(
            tmp.obs,
            obs,
            left_index=True,
            right_index=True,
            how="left",
        )
        return tmp

    @d.get_sections(base="tmk_thresh", sections=["Parameters"])
    def _sparsify_couplings(
        self,
        couplings: Dict[Key_t, AnnData],
        threshold: Threshold_t,
        copy: bool = False,
    ) -> Optional[Mapping[Key_t, AnnData]]:
        """Remove small non-zero values from :attr:`transition_matrix`.

        .. warning::
            Only dense :attr:`~anndata.AnnData.X` will be sparsified.

        Parameters
        ----------
        couplings
            Optimal transport couplings to sparsify.
        threshold
            How to remove small non-zero values from the transition matrix. Valid options are:

            - ``'auto'`` - find the maximum threshold value which will not remove every non-zero value from any row.
            - ``'auto_local'`` - same as above, but done for each transport separately.
            - :class:`float` - value in :math:`[0, 100]` corresponding to a percentage of non-zeros to remove in
              the couplings.

            Rows where all values are removed will have a uniform distribution and a warning will be issued.
        copy
            Whether to return a copy of the ``couplings`` or modify in-place.

        Returns
        -------
        If ``copy = True``, returns sparsified couplings. Otherwise, modifies ``couplings`` in-place.
        """
        if threshold == "auto":
            thresh = min(adata.X[i].max() for adata in couplings.values() for i in range(adata.n_obs))
            logg.info(f"Using automatic `threshold={thresh}`")
        elif threshold == "auto_local":
            logg.info("Using automatic `threshold` for each coupling separately")
        elif not (0 <= threshold <= 100):
            raise ValueError(f"Expected `threshold` to be in `[0, 100]`, found `{threshold}`.`")

        for key, adata in couplings.items():
            if sp.issparse(adata.X):
                continue
            if copy:
                adata = adata.copy()
            tmat = adata.X

            if threshold == "auto_local":
                thresh = min(tmat[i].max() for i in range(tmat.shape[0]))
                logg.debug(f"Using `threshold={thresh}` at `{key}`")
            elif isinstance(threshold, (int, float, np.number)):
                thresh = np.percentile(tmat.data, threshold)
                logg.debug(f"Using `threshold={thresh}` at `{key}`")

            tmat = sp.csr_matrix(tmat, dtype=tmat.dtype)
            tmat.data[tmat.data < thresh] = 0.0
            tmat.eliminate_zeros()
            couplings[key] = AnnData(tmat, obs=adata.obs, var=adata.var)

        return couplings if copy else None

    def _validate_couplings(
        self,
        couplings: Mapping[Key_t, AnnData],
    ) -> None:
        """Validate that transport maps conform to various invariants.

        Parameters
        ----------
        couplings
            Optimal transport couplings.

        Returns
        -------
        Possibly reordered transport maps.
        """

        def assert_same(expected: Sequence[Any], actual: Sequence[Any], msg: Optional[str] = None) -> None:
            try:
                pd.testing.assert_series_equal(
                    pd.Series(sorted(expected)),
                    pd.Series(sorted(actual)),
                    check_names=False,
                    check_flags=False,
                    check_index=True,
                )
            except AssertionError as e:
                raise IndexError(msg) from e

        for (src, tgt), coupling in couplings.items():
            src_obs = self.adata.obs_names[self.time == src]
            tgt_obs = self.adata.obs_names[self.time == tgt]
            assert_same(
                src_obs,
                coupling.obs_names,
                msg=f"Source observations for `{src, tgt}` don't match with `adata.obs_names`.",
            )
            assert_same(
                tgt_obs,
                coupling.var_names,
                msg=f"Source observations for `{src, tgt}` don't match with `adata.var_names`.",
            )

        for key, coupling in couplings.items():
            tmat = coupling.X
            zeros_mask = np.where(np.asarray(tmat.sum(1)).squeeze() == 0)[0]
            if np.any(zeros_mask):
                logg.warning(
                    f"Coupling at `{key}` contains `{len(zeros_mask)}` empty row(s), e.g., due to "
                    f"unbalancedness or 0s in the source marginals. Using uniform distribution"
                )
                tmat[zeros_mask] = np.ones((len(zeros_mask), tmat.shape[1]), dtype=tmat.dtype) / tmat.shape[1]

    def _coupling_to_adata(self, src: Any, tgt: Any, coupling: Coupling_t) -> AnnData:
        """Convert the coupling to :class:`~anndata.AnnData`."""
        if not isinstance(coupling, AnnData):
            coupling = AnnData(X=coupling)
            coupling.obs_names = np.asarray(self.adata.obs_names)[self.time == src]
            coupling.var_names = np.asarray(self.adata.obs_names)[self.time == tgt]

        if sp.issparse(coupling.X) and not sp.isspmatrix_csr(coupling.X):
            coupling.X = coupling.X.tocsr()

        return coupling

    def _get_default_coupling(
        self,
        couplings: Optional[Mapping[Key_t, Optional[Coupling_t]]],
        policy: Literal["sequential", "triu"],
    ) -> Tuple[Dict[Key_t, Optional[Coupling_t]], Any]:
        cats = self._time.cat.categories
        if policy == "sequential":
            if couplings is None:
                couplings = {(src, tgt): None for src, tgt in zip(cats[:-1], cats[1:])}
            reference = sorted(k for ks in couplings for k in ks)[-1]
        elif policy == "triu":
            if couplings is None:
                couplings = {(src, tgt): None for src, tgt in itertools.product(cats, cats) if src < tgt}
            reference = sorted(k for ks in couplings for k in ks)[-1]
        else:
            raise NotImplementedError(f"Handling `{policy}` is not yet implemented.")

        for keys in couplings:
            for key in keys:
                if key not in cats:
                    raise ValueError(f"Key `{key}` is not in `{sorted(cats)}`.")

        return dict(couplings), reference

    @property
    def time(self) -> pd.Series:
        """Experimental time."""
        return self._time

    @property
    def couplings(self) -> Optional[Dict[Key_t, AnnData]]:
        """Optimal transport couplings."""
        return self._couplings

    @property
    def obs(self) -> Optional[pd.DataFrame]:
        """Cell-level metadata."""
        return self._obs


def _compute_connectivity_tmat(
    adata: AnnData, density_normalize: bool = True, **kwargs: Any
) -> Union[np.ndarray, sp.spmatrix]:
    from cellrank.kernels import ConnectivityKernel

    sc.pp.neighbors(adata, **kwargs)
    return ConnectivityKernel(adata).compute_transition_matrix(density_normalize=density_normalize).transition_matrix



import contextlib
import pathlib
import types
from typing import Any, Dict, Literal, Mapping, Optional, Sequence, Tuple, Union

import scvelo as scv

import numpy as np
import pandas as pd

import matplotlib.pyplot as plt
from matplotlib import patheffects, rc_context
from matplotlib.axes import Axes
from matplotlib.patches import ArrowStyle

import scanpy as sc
from anndata import AnnData, Raw

from cellrank import logging as logg
from cellrank._utils._colors import _create_categorical_colors
from cellrank._utils._docs import d, inject_docs
from cellrank._utils._key import Key
from cellrank._utils._lineage import Lineage
from cellrank._utils._utils import RandomKeys, TestMethod, _correlation_test, save_fig
from cellrank.estimators.mixins._fate_probabilities import FateProbsMixin
from cellrank.estimators.mixins._utils import BaseProtocol, SafeGetter, logger, shadow

__all__ = ["LinDriversMixin"]


class LinDriversProtocol(BaseProtocol):
    def _write_lineage_drivers(
        self,
        drivers: Optional[pd.DataFrame],
        use_raw: bool,
        params: Mapping[str, Any] = types.MappingProxyType({}),
    ) -> None:
        ...

    @property
    def eigendecomposition(self) -> Dict[str, Any]:
        ...

    @property
    def fate_probabilities(self) -> Optional[Lineage]:
        ...

    @property
    def lineage_drivers(self) -> Optional[pd.DataFrame]:
        ...


class LinDriversMixin(FateProbsMixin):
    """Mixin that computes potential driver genes for lineages."""

    def __init__(self, **kwargs: Any):
        super().__init__(**kwargs)
        self._lineage_drivers: Optional[pd.DataFrame] = None

    @d.get_full_description(base="lineage_drivers")
    @d.get_sections(base="lineage_drivers", sections=["Parameters", "Returns"])
    @d.dedent
    @inject_docs(tm=TestMethod)
    def compute_lineage_drivers(
        self: LinDriversProtocol,
        lineages: Optional[Union[str, Sequence]] = None,
        method: Literal["fisher", "perm_test"] = TestMethod.FISHER,
        cluster_key: Optional[str] = None,
        clusters: Optional[Union[str, Sequence]] = None,
        layer: Optional[str] = None,
        use_raw: bool = False,
        confidence_level: float = 0.95,
        n_perms: int = 1000,
        seed: Optional[int] = None,
        **kwargs: Any,
    ) -> pd.DataFrame:
        """Compute driver genes per lineage.

        Correlates gene expression with lineage probabilities, for a given lineage and set of clusters.
        Often, it makes sense to restrict this to a set of clusters which are relevant for the specified lineages.

        Parameters
        ----------
        lineages
            Lineage names from :attr:`fate_probabilities`. If :obj:`None`, use all lineages.
        method
            Mode to use when calculating p-values and confidence intervals. Valid options are:

            - ``{tm.FISHER!r}`` - Fisher transformation :cite:`fisher:21`.
            - ``{tm.PERM_TEST!r}`` - permutation test.
        cluster_key
            Key in :attr:`~anndata.AnnData.obs` to obtain cluster annotations. These are considered for ``clusters``.
        clusters
            Restrict the correlations to these clusters.
        layer
            Key from :attr:`~anndata.AnnData.layers` from which to get the expression.
            If :obj:`None` or `'X'`, use :attr:`~anndata.AnnData.X`.
        use_raw
            Whether to use :attr:`~anndata.AnnData.raw` to correlate gene expression.
        confidence_level
            Confidence level for the confidence interval calculation. Must be in interval :math:`[0, 1]`.
        n_perms
            Number of permutations to use when ``method = {tm.PERM_TEST!r}``.
        seed
            Random seed when ``method = {tm.PERM_TEST!r}``.
        %(parallel)s
        kwargs
            Keyword for the correlation test.

        Returns
        -------
        %(correlation_test.returns)s Also updates the following field:

        - :attr:`lineage_drivers` - the same :class:`pandas.DataFrame` as described above.
        """
        # check that lineage probs have been computed
        method = TestMethod(method)
        fate_probs = self.fate_probabilities
        if fate_probs is None:
            raise RuntimeError("Compute `.fate_probabilities` first as `.compute_fate_probabilities()`.")

        if fate_probs.shape[1] == 1:
            logg.warning("There is only 1 lineage present. Using stationary distribution instead")
            stat_dist = self.eigendecomposition.get("stationary_dist", None)
            if stat_dist is None:
                raise RuntimeError("No stationary distribution found in `.eigendecomposition['stationary_dist']`.")
            fate_probs = Lineage(
                stat_dist,
                names=fate_probs.names,
                colors=fate_probs.colors,
            )

        # check all lin_keys exist in self.lin_names
        if isinstance(lineages, str):
            lineages = [lineages]
        if lineages is not None:
            _ = fate_probs[lineages]
        else:
            lineages = fate_probs.names
        lineages = list(lineages)

        if not len(lineages):
            raise ValueError("No lineages have been selected.")

        # use `cluster_key` and clusters to subset the data
        if clusters is not None:
            if cluster_key not in self.adata.obs.keys():
                raise KeyError(f"Unable to find clusters in `adata.obs[{cluster_key!r}]`.")
            if isinstance(clusters, str):
                clusters = [clusters]

            all_clusters = np.array(self.adata.obs[cluster_key].cat.categories)
            cluster_mask = np.array([name not in all_clusters for name in clusters])

            if any(cluster_mask):
                raise KeyError(
                    f"Clusters `{list(np.array(clusters)[cluster_mask])}` not found in "
                    f"`adata.obs[{cluster_key!r}]`."
                )
            subset_mask = np.in1d(self.adata.obs[cluster_key], clusters)
            adata_comp = self.adata[subset_mask]
            lin_probs = fate_probs[subset_mask, :]
        else:
            adata_comp = self.adata
            lin_probs = fate_probs

        # check that the layer exists, and that use raw is only used with layer X
        if layer in (None, "X"):
            if use_raw and self.adata.raw is None:
                logg.warning("No raw attribute set. Using `.X` instead")
                use_raw = False
            data = adata_comp.raw.X if use_raw else adata_comp.X
            var_names = adata_comp.raw.var_names if use_raw else adata_comp.var_names
        else:
            if layer not in self.adata.layers:
                raise KeyError(f"Layer `{layer!r}` not found in `adata.layers`.")
            if use_raw:
                logg.warning("If `use_raw=True`, layer must be `None`.")
                use_raw = False
            data = adata_comp.layers[layer]
            var_names = adata_comp.var_names

        start = logg.debug(
            f"Computing correlations for lineages `{sorted(lineages)}` restricted to clusters `{clusters}` in "
            f"layer `{'X' if layer is None else layer}` with `use_raw={use_raw}`"
        )

        lin_probs = lin_probs[lineages]
        drivers = _correlation_test(
            data,
            lin_probs,
            gene_names=var_names,
            method=method,
            n_perms=n_perms,
            seed=seed,
            confidence_level=confidence_level,
            **kwargs,
        )
        params = self._create_params()
        self._write_lineage_drivers(drivers.loc[var_names], use_raw=use_raw, params=params, time=start)

        return drivers

    @d.get_sections(base="plot_lineage_drivers", sections=["Parameters"])
    @d.dedent
    def plot_lineage_drivers(
        self: LinDriversProtocol,
        lineage: str,
        n_genes: int = 8,
        use_raw: bool = False,
        ascending: bool = False,
        ncols: Optional[int] = None,
        title_fmt: str = "{gene} qval={qval:.4e}",
        figsize: Optional[Tuple[float, float]] = None,
        dpi: Optional[int] = None,
        save: Optional[Union[str, pathlib.Path]] = None,
        **kwargs: Any,
    ) -> None:
        """Plot lineage drivers.

        Parameters
        ----------
        lineage
            Lineage for which to plot the driver genes.
        n_genes
            Top most correlated genes to plot.
        use_raw
            Whether to access data in :attr:`~anndata.AnnData.raw` or not.
        ascending
            Whether to sort the genes in ascending order.
        ncols
            Number of columns.
        title_fmt
            Title format. Can include ``{gene}``, ``{pval}``, ``{qval}`` or ``{corr}``, which will be substituted
            with the actual values.
        %(plotting)s
        kwargs
            Keyword arguments for :func:`~scvelo.pl.scatter`.

        Returns
        -------
        %(just_plots)s
        """

        def prepare_format(
            gene: str,
            *,
            pval: Optional[float],
            qval: Optional[float],
            corr: Optional[float],
        ) -> Dict[str, Any]:
            kwargs: Dict[str, Optional[Union[str, float]]] = {}
            if "{gene" in title_fmt:
                kwargs["gene"] = gene
            if "{pval" in title_fmt:
                kwargs["pval"] = float(pval) if pval is not None else np.nan
            if "{qval" in title_fmt:
                kwargs["qval"] = float(qval) if qval is not None else np.nan
            if "{corr" in title_fmt:
                kwargs["corr"] = float(corr) if corr is not None else np.nan

            return kwargs

        lin_drivers = self.lineage_drivers
        if lin_drivers is None:
            raise RuntimeError("Compute `.lineage_drivers` first as `.compute_lineage_drivers()`.")

        key = f"{lineage}_corr"
        if key not in lin_drivers:
            raise KeyError(f"Lineage `{key!r}` not found in `{list(lin_drivers.columns)}`.")

        if n_genes <= 0:
            raise ValueError(f"Expected `n_genes` to be positive, found `{n_genes}`.")

        _ = kwargs.pop("save", None)
        genes = lin_drivers.sort_values(by=key, ascending=ascending).head(n_genes)

        ncols = 4 if ncols is None else ncols
        nrows = int(np.ceil(len(genes) / ncols))

        fig, axes = plt.subplots(
            ncols=ncols,
            nrows=nrows,
            dpi=dpi,
            figsize=(ncols * 6, nrows * 4) if figsize is None else figsize,
        )
        axes = np.ravel([axes])

        _i = 0
        for _i, (gene, ax) in enumerate(zip(genes.index, axes)):
            data = genes.loc[gene]
            scv.pl.scatter(
                self.adata,
                color=gene,
                ncols=ncols,
                use_raw=use_raw,
                ax=ax,
                show=False,
                title=title_fmt.format(
                    **prepare_format(
                        gene,
                        pval=data.get(f"{lineage}_pval", None),
                        qval=data.get(f"{lineage}_qval", None),
                        corr=data.get(f"{lineage}_corr", None),
                    )
                ),
                **kwargs,
            )

        for j in range(_i + 1, len(axes)):
            axes[j].remove()

        if save is not None:
            save_fig(fig, save)

    @d.dedent
    def plot_lineage_drivers_correlation(
        self: LinDriversProtocol,
        lineage_x: str,
        lineage_y: str,
        color: Optional[str] = None,
        gene_sets: Optional[Dict[str, Sequence[str]]] = None,
        gene_sets_colors: Optional[Sequence[str]] = None,
        use_raw: bool = False,
        cmap: str = "RdYlBu_r",
        fontsize: int = 12,
        adjust_text: bool = False,
        legend_loc: Optional[str] = "best",
        figsize: Optional[Tuple[float, float]] = (4, 4),
        dpi: Optional[int] = None,
        save: Optional[Union[str, pathlib.Path]] = None,
        show: bool = True,
        **kwargs: Any,
    ) -> Optional[Axes]:
        """Show scatter plot of gene-correlations between two lineages.

        Optionally, a :class:`dict` of gene names can be passed to highlight in the plot.

        Parameters
        ----------
        lineage_x
            Name of the lineage on the x-axis.
        lineage_y
            Name of the lineage on the y-axis.
        color
            Key in :attr:`~anndata.AnnData.var` or :attr:`~anndata.AnnData.varm`, preferring for the former.
        gene_sets
            Gene sets annotations of the form ``{'gene_set_name': ['gene_1', 'gene_2'], ...}``.
        gene_sets_colors
            List of colors where each entry corresponds to a gene set from ``genes_sets``.
            If `None` and keys in ``gene_sets`` correspond to lineage names, use the lineage colors.
            Otherwise, use default colors.
        use_raw
            Whether to access :attr:`~anndata.AnnData.raw` or not.
        cmap
            Colormap to use.
        fontsize
            Size of the text when plotting ``gene_sets``.
        adjust_text
            Whether to automatically adjust text in order to reduce overlap.
        legend_loc
            Position of the legend. If :obj:`None`, don't show the legend. Only used when ``gene_sets != None``.
        %(plotting)s
        show
            If :obj:`False`, return the :class:`~matplotlib.axes.Axes` object.
        kwargs
            Keyword arguments for :func:`~scanpy.pl.scatter`.

        Returns
        -------
        If ``show = True``, nothing, just plots, otherwise returns the axes object.
        Optionally saves it based on ``save``.

        Notes
        -----
        This plot is based on the following
        `notebook <https://github.com/theislab/gastrulation_analysis/blob/main/6_cellrank.ipynb>`_ by *Maren Büttner*.
        """
        from cellrank.pl._utils import _position_legend

        if use_raw and self.adata.raw is None:
            logg.warning("No raw attribute set. Setting `use_raw=False`")
            use_raw = False
        adata = self.adata.raw if use_raw else self.adata
        dkey = Key.varm.lineage_drivers(self.backward)
        # this makes a useless copy, wish there was another way
        if isinstance(adata, Raw):
            adata = adata.to_adata()

        if color is not None:
            if not isinstance(color, str):
                raise TypeError(
                    "Expected `color` to be either of type `str` or `None`, " f"found `{type(color).__name__}`."
                )
            if color not in adata.var and color not in adata.varm[dkey]:
                raise KeyError(
                    f"Unable to find key `{color!r}` in `adata{'.raw' if use_raw else ''}.var` "
                    f"or `adata.{'.raw' if use_raw else ''}varm[{dkey!r}]`."
                )

        # silent assumption: `.compute_lineage_drivers()` always writes to AnnData
        key1, key2 = f"{lineage_x}_corr", f"{lineage_y}_corr"
        if key1 not in adata.varm[dkey] or key2 not in adata.varm[dkey]:
            haystack = "adata.raw.varm" if use_raw else "adata.varm"
            raise RuntimeError(
                f"Unable to find correlations in `{haystack}[{key1!r}]` or `{haystack}[{key2!r}]`."
                f"Compute `.lineage_drivers` first as "
                f"`.compute_lineage_drivers([{lineage_x!r}, {lineage_y!r}], use_raw={use_raw})`."
            )
        # produce the actual scatter plot
        ctx = {"figure.figsize": figsize, "figure.dpi": dpi}
        for key in list(ctx.keys()):
            if ctx[key] is None:
                del ctx[key]
        with rc_context(ctx), RandomKeys(adata, n=3, where="var") as keys:
            adata.var[keys[0]] = adata.varm[dkey][key1]
            adata.var[keys[1]] = adata.varm[dkey][key2]
            if color is not None:
                if color in adata.var:
                    ckey = color
                else:
                    adata.var[keys[2]] = adata.varm[dkey][color]
                    ckey = keys[2]

            ax = sc.pl.scatter(
                adata,
                x=keys[0],
                y=keys[1],
                color_map=cmap,
                use_raw=False,
                show=False,
                color=None if color is None else ckey,
                **kwargs,
            )

            # add some lines to highlight the origin
            xmin, xmax = np.nanmin(adata.var[keys[0]]), np.nanmax(adata.var[keys[0]])
            ymin, ymax = np.nanmin(adata.var[keys[1]]), np.nanmax(adata.var[keys[1]])
            ax.hlines(0, xmin=xmin, xmax=xmax, color="grey", alpha=0.5, zorder=-1)
            ax.vlines(0, ymin=ymin, ymax=ymax, color="grey", alpha=0.5, zorder=-1)

        fig = ax.figure
        ax.set_xlabel(key1)
        ax.set_ylabel(key2)
        ax.set_title(color)

        # annotate the passed set of genes
        if gene_sets is not None:
            if gene_sets_colors is None:
                try:
                    # fmt: off
                    sets = list(gene_sets.keys())
                    gene_sets_colors = self.fate_probabilities[sets].colors
                    # fmt: on
                except KeyError:
                    logg.warning("Unable to determine gene sets colors from lineages. Using default colors")
                    gene_sets_colors = _create_categorical_colors(len(gene_sets))
            if len(gene_sets_colors) != len(gene_sets):
                raise ValueError(
                    f"Expected `gene_sets_colors` to be of length `{len(gene_sets)}`, "
                    f"found `{len(gene_sets_colors)}`."
                )

            path_effect = [
                patheffects.Stroke(linewidth=2, foreground="w", alpha=0.8),
                patheffects.Normal(),
            ]
            annots = []
            for (key, values), color in zip(gene_sets.items(), gene_sets_colors):
                arrowprops = (
                    {
                        "arrowstyle": ArrowStyle.CurveFilledB(head_width=0.1, head_length=0.2),
                        "fc": color,
                        "ec": color,
                    }
                    if adjust_text
                    else None
                )
                values = set(values) & set(adata.var_names)
                for value in values:
                    x = adata.varm[dkey].loc[value, key1]
                    y = adata.varm[dkey].loc[value, key2]

                    annot = ax.annotate(
                        value,
                        xy=(x, y),
                        va="top",
                        ha="left",
                        path_effects=path_effect,
                        arrowprops=arrowprops,
                        size=fontsize,
                        c=color,
                    )
                    annots.append(annot)
                if values:
                    ax.scatter([], [], color=color, label=key)

            if adjust_text:
                try:
                    import adjustText

                    start = logg.info("Adjusting text position")
                    adjustText.adjust_text(
                        annots,
                        # https://github.com/theislab/cellrank/issues/1033
                        x=np.array(adata.varm[dkey][key1], copy=True),
                        y=np.array(adata.varm[dkey][key2], copy=True),
                        ax=ax,
                    )
                    logg.info("    Finish", time=start)
                except ImportError:
                    logg.error("Please install `adjustText` first as `pip install adjustText`")
            if len(annots) and legend_loc not in (None, "none"):
                _position_legend(ax, legend_loc=legend_loc)

        if save is not None:
            save_fig(fig, path=save)

        if not show:
            return ax

    @logger
    @shadow
    def _write_fate_probabilities(
        self: LinDriversProtocol,
        fate_probs: Optional[Lineage],
        params: Mapping[str, Any] = types.MappingProxyType({}),
    ) -> str:
        self._write_lineage_drivers(None, use_raw=False, log=False)
        with contextlib.suppress(AttributeError):
            self._write_lineage_drivers(None, use_raw=True, log=False)

        return super()._write_fate_probabilities(fate_probs, params=params, log=False)

    @logger
    @shadow
    def _write_lineage_drivers(
        self: LinDriversProtocol,
        drivers: Optional[pd.DataFrame],
        use_raw: bool,
        params: Mapping[str, Any] = types.MappingProxyType({}),
    ) -> str:
        # fmt: off
        key = Key.varm.lineage_drivers(self.backward)
        self._set("_lineage_drivers", self.adata.raw.varm if use_raw else self.adata.varm, key=key, value=drivers)
        self.params[key] = dict(params)
        # fmt: on

        return f"Adding `adata.{'raw.' if use_raw else ''}varm[{key!r}]`\n" f"       `.lineage_drivers`\n" "    Finish"

    def _read_fate_probabilities(self: LinDriversProtocol, adata: AnnData) -> bool:
        ok = super()._read_fate_probabilities(adata)
        if not ok:
            return False

        key = Key.varm.lineage_drivers(self.backward)
        with SafeGetter(self, allowed=(KeyError, AttributeError)) as sg:
            self.params[key] = self._read_params(key)
            if self.params[key].get("use_raw", False):
                self._lineage_drivers = self._get(
                    obj=self.adata.raw.varm,
                    key=key,
                    shadow_attr="varm",
                    dtype=pd.DataFrame,
                )
            else:
                self._lineage_drivers = self._get(obj=self.adata.varm, key=key, shadow_attr="varm", dtype=pd.DataFrame)

        return sg.ok

    @property
    @d.dedent
    def lineage_drivers(self) -> Optional[pd.DataFrame]:
        """Potential lineage drivers.

        Computes Pearson correlation of each gene with fate probabilities for every terminal state. High Pearson
        correlation indicates potential lineage drivers. Also computes p-values and confidence intervals.

        Returns
        -------
        %(correlation_test.returns)s
        """
        return self._lineage_drivers



from cellrank.estimators import mixins
from cellrank.estimators._base_estimator import BaseEstimator
from cellrank.estimators.terminal_states import CFLARE, GPCCA, TermStatesEstimator

__all__ = ["mixins", "BaseEstimator", "TermStatesEstimator", "CFLARE", "GPCCA"]



import collections
import copy
import enum
import types
import warnings
from typing import Any, Literal, Mapping, Optional, Union

from pygam import GAM as pGAM
from pygam import (
    ExpectileGAM,
    GammaGAM,
    InvGaussGAM,
    LinearGAM,
    LogisticGAM,
    PoissonGAM,
    s,
)

import numpy as np

from anndata import AnnData

from cellrank import logging as logg
from cellrank._utils._docs import d
from cellrank._utils._enum import ModeEnum
from cellrank._utils._utils import _filter_kwargs
from cellrank.models import BaseModel

__all__ = ["GAM"]


class GamLinkFunction(ModeEnum):
    IDENTITY = enum.auto()
    LOGIT = enum.auto()
    INVERSE = enum.auto()
    LOG = enum.auto()
    INV_SQUARED = enum.auto()


class GamDistribution(ModeEnum):
    NORMAL = enum.auto()
    BINOMIAL = enum.auto()
    POISSON = enum.auto()
    GAMMA = enum.auto()
    GAUSSIAN = enum.auto()
    INV_GAUSS = enum.auto()


_gams = collections.defaultdict(
    lambda: pGAM,
    {
        (GamDistribution.NORMAL, GamLinkFunction.IDENTITY): LinearGAM,
        (GamDistribution.BINOMIAL, GamLinkFunction.LOGIT): LogisticGAM,
        (GamDistribution.POISSON, GamLinkFunction.LOG): PoissonGAM,
        (GamDistribution.GAMMA, GamLinkFunction.LOG): GammaGAM,
        (GamDistribution.INV_GAUSS, GamLinkFunction.LOG): InvGaussGAM,
    },
)


@d.dedent
class GAM(BaseModel):
    """Fit Generalized Additive Models (GAMs) using :mod:`pygam`.

    Parameters
    ----------
    %(adata)s
    n_knots
        Number of knots.
    spline_order
        Order of the splines, e.g., :math:`3` for cubic splines.
    distribution
        Name of the distribution. Available distributions can be found
        `here <https://pygam.readthedocs.io/en/latest/notebooks/tour_of_pygam.html#Distribution:>`__.
    link
        Name of the link function. Available link functions can be found
        `here <https://pygam.readthedocs.io/en/latest/notebooks/tour_of_pygam.html#Link-function:>`__.
    max_iter
        Maximum number of iterations for optimization.
    expectile
        Expectile for :class:`~pygam.pygam.ExpectileGAM`. This forces the distribution to be ``'normal'``
        and link function to ``'identity'``. Must be in :math:`(0, 1)`.
    grid
        Whether to perform a grid search. Keys correspond to a parameter names and values to range to be searched.
        If ``'default'``, use the default grid. If :obj:`None`, don't perform a grid search.
    spline_kwargs
        Keyword arguments for :func:`~pygam.terms.s`.
    kwargs
        Keyword arguments for the :class:`~pygam.pygam.GAM`.
    """

    def __init__(
        self,
        adata: AnnData,
        n_knots: Optional[int] = 6,
        spline_order: int = 3,
        distribution: Literal["normal", "binomial", "poisson", "gamma", "gaussian", "inv_gauss"] = "gamma",
        link: Literal["identity", "logit", "inverse", "log", "inv_squared"] = "log",
        max_iter: int = 2000,
        expectile: Optional[float] = None,
        grid: Optional[Union[str, Mapping[str, Any]]] = None,
        spline_kwargs: Mapping[str, Any] = types.MappingProxyType({}),
        **kwargs: Any,
    ):
        term = s(
            0,
            spline_order=spline_order,
            n_splines=n_knots,
            **_filter_kwargs(s, **{**{"lam": 3, "penalties": ["derivative", "l2"]}, **spline_kwargs}),
        )
        link = GamLinkFunction(link)
        distribution = GamDistribution(distribution)
        if distribution == GamDistribution.GAUSSIAN:
            distribution = GamDistribution.NORMAL

        if expectile is not None:
            if not (0 < expectile < 1):
                raise ValueError(f"Expected `expectile` to be in `(0, 1)`, found `{expectile}`.")
            if distribution != "normal" or link != "identity":
                logg.warning(
                    f"Expectile GAM works only with `normal` distribution and `identity` link function,"
                    f"found `{distribution!r}` distribution and {link!r} link functions."
                )
            model = ExpectileGAM(term, expectile=expectile, max_iter=max_iter, verbose=False, **kwargs)
        else:
            # doing it like this ensure that user can specify scale
            gam = _gams[distribution, link]

            filtered_kwargs = _filter_kwargs(gam.__init__, **kwargs)
            if len(kwargs) != len(filtered_kwargs):
                raise TypeError(
                    f"Invalid arguments `{list(set(kwargs) - set(filtered_kwargs))}` " f"for `{type(gam).__name__!r}`."
                )

            filtered_kwargs["link"] = link
            filtered_kwargs["distribution"] = distribution

            model = gam(
                term,
                max_iter=max_iter,
                verbose=False,
                **_filter_kwargs(gam.__init__, **filtered_kwargs),
            )
        super().__init__(adata, model=model)

        if grid is None:
            self._grid = None
        elif isinstance(grid, dict):
            self._grid = copy.copy(grid)
        elif isinstance(grid, str):
            self._grid = object() if grid == "default" else None
        else:
            raise TypeError(f"Expected `grid` to be `dict`, `str` or `None`, found `{type(grid).__name__!r}`.")

    @d.dedent
    def fit(
        self,
        x: Optional[np.ndarray] = None,
        y: Optional[np.ndarray] = None,
        w: Optional[np.ndarray] = None,
        **kwargs: Any,
    ) -> "GAM":
        """%(base_model_fit.full_desc)s

        Parameters
        ----------
        %(base_model_fit.parameters)s

        Returns
        -------
        Fits the model and returns self.
        """  # noqa

        super().fit(x, y, w, **kwargs)

        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                category=DeprecationWarning,
                message=".* is a deprecated alias for the builtin",
            )
            if self._grid is not None:
                # use default search
                grid = {} if not isinstance(self._grid, dict) else self._grid
                try:
                    self.model.gridsearch(
                        self.x,
                        self.y,
                        weights=self.w,
                        keep_best=True,
                        progress=False,
                        **grid,
                        **kwargs,
                    )
                    return self
                except Exception as e:  # noqa: BLE001
                    # workaround for: https://github.com/dswah/pyGAM/issues/273
                    self.model.fit(self.x, self.y, weights=self.w, **kwargs)
                    logg.error(f"Grid search failed, reason: `{e}`. Fitting with default values")

            try:
                self.model.fit(self.x, self.y, weights=self.w, **kwargs)
                return self
            except Exception as e:  # noqa: BLE001
                raise RuntimeError(
                    f"Unable to fit `{type(self).__name__}` for gene "
                    f"`{self._gene!r}` in lineage `{self._lineage!r}`. Reason: `{e}`"
                ) from e

    @d.dedent
    def predict(
        self,
        x_test: Optional[np.ndarray] = None,
        key_added: Optional[str] = "_x_test",
        **kwargs: Any,
    ) -> np.ndarray:
        """%(base_model_predict.full_desc)s

        Parameters
        ----------
        %(base_model_predict.parameters)s

        Returns
        -------
        %(base_model_predict.returns)s
        """  # noqa

        x_test = self._check(key_added, x_test)

        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                category=DeprecationWarning,
                message=".* is a deprecated alias for the builtin",
            )
            self._y_test = self.model.predict(x_test, **kwargs)
        self._y_test = np.squeeze(self._y_test).astype(self._dtype)

        return self.y_test

    @d.dedent
    def confidence_interval(self, x_test: Optional[np.ndarray] = None, **kwargs: Any) -> np.ndarray:
        """%(base_model_ci.summary)s

        Parameters
        ----------
        %(base_model_ci.parameters)s

        Returns
        -------
        %(base_model_ci.returns)s
        """  # noqa

        x_test = self._check("_x_test", x_test)
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                category=DeprecationWarning,
                message=".* is a deprecated alias for the builtin",
            )
            self._conf_int = self.model.confidence_intervals(x_test, **kwargs).astype(self._dtype)

        return self.conf_int

    @d.dedent
    def copy(self) -> "BaseModel":
        """%(copy)s"""  # noqa
        res = GAM(self.adata)
        self._shallowcopy_attributes(res)
        res._grid = copy.deepcopy(self._grid)
        return res



import contextlib
import io
import pathlib
import types
from typing import Any, Dict, Literal, Mapping, Optional, Tuple, Union

from pygpcca import GPCCA
from pygpcca._sorted_schur import _check_conj_split

import numpy as np
import scipy.sparse as sp

import matplotlib.pyplot as plt
import seaborn as sns
from mpl_toolkits.axes_grid1 import make_axes_locatable

from anndata import AnnData

from cellrank import logging as logg
from cellrank._utils._docs import d
from cellrank._utils._key import Key
from cellrank._utils._utils import _eigengap, save_fig
from cellrank.estimators.mixins._utils import BaseProtocol, SafeGetter, logger, shadow

__all__ = ["SchurMixin"]

EPS = np.finfo(np.float64).eps


class SchurProtocol(BaseProtocol):
    _gpcca: Optional[GPCCA] = None
    _invalid_n_states: Optional[np.ndarray] = None

    schur_vectors: Optional[np.ndarray] = None
    _schur_matrix: Optional[np.ndarray] = None
    _eigendecomposition: Optional[Dict[str, Any]] = None

    def transition_matrix(self) -> Union[np.ndarray, sp.spmatrix]:
        ...

    def _write_schur_decomposition(
        self,
        decomp: Dict[str, Any],
        vectors: np.ndarray,
        matrix: np.ndarray,
        params: Mapping[str, Any] = types.MappingProxyType({}),
    ) -> str:
        ...


class SchurMixin:
    """Mixin that computes Schur decomposition."""

    def __init__(self, **kwargs: Any):
        super().__init__(**kwargs)
        self._gpcca: Optional[GPCCA] = None
        self._invalid_n_states: Optional[np.ndarray] = None

        self._schur_vectors: Optional[np.ndarray] = None
        self._schur_matrix: Optional[np.ndarray] = None
        self._eigendecomposition: Optional[Dict[str, Any]] = None

    @property
    @d.get_summary(base="schur_vectors")
    @d.get_extended_summary(base="schur_vectors")
    def schur_vectors(self) -> Optional[np.ndarray]:
        """Real Schur vectors of the transition matrix.

        The real Schur decomposition is a generalization of the Eigendecomposition and can be computed for any
        real-valued, square matrix :math:`A`. It is given by :math:`A = Q R Q^T`, where :math:`Q` contains the
        real Schur vectors and :math:`R` is the Schur matrix. :math:`Q` is orthogonal and :math:`R` is quasi-upper
        triangular with 1x1 and 2x2 blocks on the diagonal.

        If PETSc and SLEPc are installed, only the leading Schur vectors are computed.
        """
        return self._schur_vectors

    @property
    @d.get_summary(base="schur_matrix")
    @d.dedent
    def schur_matrix(self) -> Optional[np.ndarray]:
        """Schur matrix.

        %(schur_vectors.summary_ext)s
        """
        return self._schur_matrix

    @property
    @d.get_summary(base="eigen")
    def eigendecomposition(self) -> Optional[Dict[str, Any]]:
        """Eigendecomposition of the :attr:`transition_matrix`.

        For non-symmetric real matrices, left and right eigenvectors will in general be different and complex.
        We compute both left and right eigenvectors.

        Returns
        -------
        A dictionary with the following keys:

        - ``'D'`` - the eigenvalues.
        - ``'eigengap'`` - the `eigengap <https://en.wikipedia.org/wiki/Eigengap>`__.
        - ``'params'`` - parameters used for the computation.
        - ``'V_l'`` - left eigenvectors (optional).
        - ``'V_r'`` - right eigenvectors (optional).
        - ``'stationary_dist'`` - stationary distribution of the :attr:`transition_matrix`, if present.
        """
        return self._eigendecomposition

    @d.dedent
    def compute_schur(
        self: SchurProtocol,
        n_components: int = 20,
        initial_distribution: Optional[np.ndarray] = None,
        method: Literal["krylov", "brandts"] = "krylov",
        which: Literal["LR", "LM"] = "LR",
        alpha: float = 1.0,
        verbose: Optional[bool] = None,
    ) -> "SchurMixin":
        """Compute the Schur decomposition.

        Parameters
        ----------
        n_components
            Number of Schur vectors to compute.
        initial_distribution
            Input distribution over all cells. If :obj:`None`, uniform distribution is used.
        method
            Method for calculating the Schur vectors. Valid options are:

            - ``'krylov'`` - an iterative procedure that computes a partial, sorted Schur decomposition for
              large, sparse matrices.
            - ``'brandts'`` - full sorted Schur decomposition of a dense matrix.

            For benefits of each method, see :class:`~pygpcca.GPCCA`.
        %(eigen)s
        verbose
            Whether to print extra information when computing the Schur decomposition.
            If :obj:`None`, it's disabled when ``method = 'krylov'``.

        Returns
        -------
        Self and just updates the following fields:

        - :attr:`schur_vectors` - %(schur_vectors.summary)s
        - :attr:`schur_matrix` -  %(schur_matrix.summary)s
        - :attr:`eigendecomposition` - %(eigen.summary)s
        """
        from cellrank._utils._linear_solver import _is_petsc_slepc_available

        if n_components < 2:
            logg.warning(
                f"Number of Schur vectors `>=2`, but only `{n_components}` " f"were requested. Using `n_components=2`"
            )
            n_components = 2

        if method not in ("brandts", "krylov"):
            raise ValueError(f"Invalid method `{method!r}`. Valid options are:`'brandts'` or `'krylov'`.")

        if not _is_petsc_slepc_available():
            method = "brandts"
            logg.warning(f"Unable to import `petsc4py` or `slepc4py`. Using `method={method!r}`")
        if verbose is None:
            verbose = method == "brandts"

        tmat = self.transition_matrix
        if method == "brandts" and sp.issparse(self.transition_matrix):
            logg.warning("For `method='brandts'`, dense matrix is required. Densifying")
            tmat = tmat.A

        self._gpcca = GPCCA(tmat, eta=initial_distribution, z=which, method=method)
        start = logg.info("Computing Schur decomposition")

        with (contextlib.nullcontext if verbose else contextlib.redirect_stdout)(io.StringIO()):
            try:
                self._gpcca._do_schur_helper(n_components)
            except ValueError as e:
                if "will split complex conjugate eigenvalues" not in str(e):
                    raise
                logg.warning(
                    f"Using `{n_components}` components would split a block of complex conjugate eigenvalues. "
                    f"Using `n_components={n_components + 1}`"
                )
                self._gpcca._do_schur_helper(n_components + 1)

        self._invalid_n_states = np.array(
            [i for i in range(2, len(self._gpcca._p_eigenvalues)) if _check_conj_split(self._gpcca._p_eigenvalues[:i])]
        )
        if len(self._invalid_n_states):
            logg.info(f"When computing macrostates, choose a number of states NOT in `{list(self._invalid_n_states)}`")

        self._write_schur_decomposition(
            {
                "D": self._gpcca._p_eigenvalues,
                "eigengap": _eigengap(self._gpcca._p_eigenvalues, alpha),
                "params": {
                    "which": which,
                    "k": len(self._gpcca._p_eigenvalues),
                    "alpha": alpha,
                },
            },
            vectors=self._gpcca._p_X,
            matrix=self._gpcca._p_R,
            params=self._create_params(),
            time=start,
        )
        return self

    @d.dedent
    def plot_schur_matrix(
        self,
        title: Optional[str] = "schur matrix",
        cmap: str = "viridis",
        figsize: Optional[Tuple[float, float]] = None,
        dpi: Optional[float] = 80,
        save: Optional[Union[str, pathlib.Path]] = None,
        **kwargs: Any,
    ) -> None:
        """Plot the Schur matrix.

        Parameters
        ----------
        title
            Title of the figure.
        cmap
            Colormap to use.
        %(plotting)s
        kwargs
            Keyword arguments for :func:`~seaborn.heatmap`.

        Returns
        -------
        %(just_plots)s
        """
        schur_matrix = self.schur_matrix
        if schur_matrix is None:
            raise RuntimeError("Compute Schur matrix first as `.compute_schur()`.")

        fig, ax = plt.subplots(figsize=schur_matrix.shape if figsize is None else figsize, dpi=dpi)

        divider = make_axes_locatable(ax)  # square=True make the colorbar a bit bigger
        cbar_ax = divider.append_axes("right", size="2%", pad=0.1)

        mask = np.zeros_like(schur_matrix, dtype=np.bool_)
        mask[np.tril_indices_from(mask, k=-1)] = True
        mask[~np.isclose(schur_matrix, 0.0)] = False

        vmin, vmax = (
            np.min(schur_matrix[~mask]),
            np.max(schur_matrix[~mask]),
        )

        kwargs["fmt"] = kwargs.get("fmt", "0.2f")
        sns.heatmap(
            data=schur_matrix,
            cmap=cmap,
            square=True,
            annot=True,
            vmin=vmin,
            vmax=vmax,
            cbar_ax=cbar_ax,
            cbar_kws={"ticks": np.linspace(vmin, vmax, 10)},
            mask=mask,
            xticklabels=[],
            yticklabels=[],
            ax=ax,
            **kwargs,
        )
        ax.set_title(title)

        if save is not None:
            save_fig(fig, path=save)

    @logger
    @shadow
    def _write_schur_decomposition(
        self: SchurProtocol,
        decomp: Dict[str, Any],
        vectors: np.ndarray,
        matrix: np.ndarray,
        params: Mapping[str, Any] = types.MappingProxyType({}),
    ) -> str:
        key = Key.uns.schur_matrix(self.backward)
        self._set("_schur_matrix", self.adata.uns, key=key, value=matrix)
        key = Key.obsm.schur_vectors(self.backward)
        self._set("_schur_vectors", self.adata.obsm, key=key, value=vectors)
        key = Key.uns.eigen(self.backward)
        self._set("_eigendecomposition", self.adata.uns, key=key, value=decomp)
        self.params[f"schur_decomposition_{Key.backward(self.backward)}"] = dict(params)

        return (
            f"Adding `adata.uns[{key!r}]`\n"
            f"       `.schur_vectors`\n"
            f"       `.schur_matrix`\n"
            f"       `.eigendecomposition`\n"
            "    Finish"
        )

    def _read_schur_decomposition(self: SchurProtocol, adata: AnnData, allow_missing: bool = True) -> bool:
        key = Key.uns.eigen(self.backward)
        with SafeGetter(self, allowed=KeyError) as sg:
            self._eigendecomposition = self._get(
                obj=adata.uns,
                key=key,
                shadow_attr="uns",
                dtype=dict,
                allow_missing=allow_missing,
            )
            key = Key.obsm.schur_vectors(self.backward)
            self._schur_vectors = self._get(
                obj=self.adata.obsm,
                key=key,
                shadow_attr="obsm",
                dtype=np.ndarray,
                allow_missing=allow_missing,
            )
            key = Key.uns.schur_matrix(self.backward)
            self._schur_matrix = self._get(
                obj=self.adata.uns,
                key=key,
                shadow_attr="uns",
                dtype=np.ndarray,
                allow_missing=allow_missing,
            )
            key = f"schur_decomposition_{Key.backward(self.backward)}"
            self.params[key] = self._read_params(key)

        return sg.ok



import enum
from typing import Any, List, Literal, Optional, Tuple

import numpy as np
import pandas as pd
import scipy.sparse as sp
import scipy.stats as st

from anndata import AnnData

from cellrank import logging as logg
from cellrank._utils._docs import d, inject_docs
from cellrank._utils._enum import ModeEnum
from cellrank._utils._key import Key
from cellrank._utils._utils import _correlation_test
from cellrank.kernels._pseudotime_kernel import PseudotimeKernel

__all__ = ["CytoTRACEKernel"]


class CytoTRACEAggregation(ModeEnum):
    MEAN = enum.auto()
    MEDIAN = enum.auto()
    GMEAN = enum.auto()
    HMEAN = enum.auto()


@d.dedent
class CytoTRACEKernel(PseudotimeKernel):
    """Kernel which computes directed transition probabilities using the CytoTRACE score :cite:`gulati:20`.

    .. seealso::
        - See :doc:`../../../notebooks/tutorials/kernels/400_cytotrace` on how to
          compute the :attr:`~cellrank.kernels.CytoTRACEKernel.transition_matrix` based on the CytoTRACE score.

    The k-NN graph contains information about the (undirected) connectivities among cells, reflecting their similarity.
    CytoTRACE can be used to estimate cellular plasticity and in turn, a pseudotemporal ordering of cells from more
    plastic to less plastic states. It relies on the assumption that differentiated cells express, on average,
    fewer genes than naive cells.

    This kernel internally uses the :class:`~cellrank.kernels.PseudotimeKernel` to direct the k-NN graph
    on the basis of the CytoTRACE pseudotime.

    %(density_correction)s

    Parameters
    ----------
    %(adata)s
    %(backward)s
    kwargs
        Keyword arguments for the :class:`~cellrank.kernels.PseudotimeKernel`.

    Examples
    --------
    .. code-block::

        import scvelo as scv
        import cellrank as cr

        adata = cr.datasets.pancreas()

        sc.pp.filter_genes(adata, min_cells=10)
        sc.pp.normalize_total(adata)
        sc.pp.log1p(adata)
        sc.pp.highly_variable_genes(adata)

        # CytoTRACE by default uses imputed data - a simple way to compute
        # k-NN imputed data is to use scVelo's moments function.
        # However, note that this function expects `spliced` counts because
        # it's designed for RNA velocity, so we're using a simple hack here:
        if 'spliced' not in adata.layers or 'unspliced' not in adata.layers:
            adata.layers['spliced'] = adata.X
            adata.layers['unspliced'] = adata.X
        scv.pp.moments(adata)

        ctk = cr.kernels.CytoTRACEKernel(adata)
        ckt = ctk.compute_cytotrace().compute_transition_matrix()
    """

    def __init__(
        self,
        adata: AnnData,
        backward: bool = False,
        **kwargs: Any,
    ):
        _ = kwargs.pop("time_key", None)
        super().__init__(
            adata,
            backward=backward,
            time_key=Key.cytotrace("pseudotime"),
            **kwargs,
        )

    def _read_from_adata(
        self,
        time_key: str,
        **kwargs: Any,
    ) -> None:
        try:
            super()._read_from_adata(time_key=time_key, **kwargs)
        except KeyError as e:
            self._pseudotime: Optional[np.ndarray] = None
            if "Unable to find pseudotime" not in str(e):
                raise
            super(PseudotimeKernel, self)._read_from_adata(**kwargs)

    @d.get_sections(base="cytotrace", sections=["Parameters"])
    @inject_docs(ct=CytoTRACEAggregation)
    def compute_cytotrace(
        self,
        layer: Optional[str] = "Ms",
        aggregation: Literal["mean", "median", "hmean", "gmean"] = CytoTRACEAggregation.MEAN,
        use_raw: bool = False,
        n_genes: int = 200,
    ) -> "CytoTRACEKernel":
        """Re-implementation of the CytoTRACE algorithm :cite:`gulati:20` to estimate cellular plasticity.

        Computes the number of genes expressed per cell and ranks genes according to their correlation with this
        measure. Next, it selects to top-correlating genes and aggregates their (imputed) expression to obtain
        the CytoTRACE score. A high score stands for high differentiation potential (naive, plastic cells) and
        a low score stands for low differentiation potential (mature, differentiation cells).

        Parameters
        ----------
        layer
            Key in :attr:`~anndata.AnnData.layers` or ``'X'`` for :attr:`~anndata.AnnData.X`
            from where to get the expression.
        aggregation
            How to aggregate expression of the top-correlating genes. Valid options are:

            - ``{ct.MEAN!r}`` - arithmetic mean.
            - ``{ct.MEDIAN!r}`` - median.
            - ``{ct.HMEAN!r}`` - harmonic mean.
            - ``{ct.GMEAN!r}`` - geometric mean.
        use_raw
            Whether to use the :attr:`~anndata.AnnData.raw` to compute the number of genes expressed per cell
            (#genes/cell) and the correlation of gene expression across cells with #genes/cell.
        n_genes
            Number of top positively correlated genes to compute the CytoTRACE score.

        Returns
        -------
        Nothing, just modifies :attr:`~anndata.AnnData.obs` with the following keys:

        - ``'ct_score'`` - the normalized CytoTRACE score.
        - ``'ct_pseudotime'`` - associated pseudotime, essentially ``1 - CytoTRACE score``.
        - ``'ct_num_exp_genes'`` - the number of genes expressed per cell, basis of the CytoTRACE score.

        It also modifies :attr:`~anndata.AnnData.var` with the following keys:

        - ``'ct_gene_corr'`` - the correlation as specified above.
        - ``'ct_correlates'`` - indication of the genes used to compute the CytoTRACE score, i.e. the ones that
          correlated positively with ``'ct_num_exp_genes'``.

        Notes
        -----
        This will not exactly reproduce the results of the original CytoTRACE algorithm :cite:`gulati:20` because we
        allow for any normalization and imputation techniques whereas CytoTRACE has built-in specific methods for that.
        """
        from cellrank._utils import Lineage

        aggregation = CytoTRACEAggregation(aggregation)

        if use_raw and self.adata.raw is None:
            logg.warning("`adata.raw` is `None`. Setting `use_raw=False`")
            use_raw = False
        if layer not in (None, "X") and layer not in self.adata.layers:
            raise KeyError(
                f"Unable to find `{layer!r}` in `adata.layers`. "
                f"Valid option are: `{sorted({'X'} | set(self.adata.layers.keys()))}`."
            )

        adata_mraw = self.adata.raw if use_raw else self.adata
        msg = f"Computing CytoTRACE score with `{adata_mraw.n_vars}` genes"
        if adata_mraw.n_vars < 10000:
            msg += ". Consider using more than `10000` genes"
        start = logg.info(msg)

        # compute number of expressed genes per cell
        num_exp_genes = np.asarray((adata_mraw.X > 0).sum(axis=1)).squeeze()

        logg.debug("Correlating all genes with number of genes expressed per cell")
        gene_corr = _correlation_test(
            adata_mraw.X,
            Lineage(num_exp_genes[:, None], names=["gene"]),
            gene_names=adata_mraw.var_names,
        )["gene_corr"]

        cytotrace_score, pos_top_genes = self._compute_score(
            gene_corr,
            layer=layer,
            aggregation=aggregation,
            ascending=False,
            n_genes=n_genes,
        )
        cytotrace_score -= cytotrace_score.min()
        cytotrace_score /= cytotrace_score.max()

        self.adata.obs[Key.cytotrace("score")] = cytotrace_score
        self.adata.obs[Key.cytotrace("pseudotime")] = self._pseudotime = 1 - cytotrace_score
        self.adata.obs[Key.cytotrace("num_exp_genes")] = num_exp_genes
        self.adata.var[Key.cytotrace("gene_corr")] = gene_corr
        self.adata.var[Key.cytotrace("correlates")] = False
        self.adata.var.loc[pos_top_genes, Key.cytotrace("correlates")] = True
        self.adata.uns[Key.cytotrace("params")] = {
            "aggregation": str(aggregation),
            "layer": layer,
            "use_raw": use_raw,
            "n_genes": len(pos_top_genes),
        }

        logg.info(
            f"Adding `adata.obs[{Key.cytotrace('score')!r}]`\n"
            f"       `adata.obs[{Key.cytotrace('pseudotime')!r}]`\n"
            f"       `adata.obs[{Key.cytotrace('num_exp_genes')!r}]`\n"
            f"       `adata.var[{Key.cytotrace('gene_corr')!r}]`\n"
            f"       `adata.var[{Key.cytotrace('correlates')!r}]`\n"
            f"       `adata.uns[{Key.cytotrace('params')!r}]`\n"
            f"    Finish",
            time=start,
        )

        return self

    def _compute_score(
        self,
        gene_corr: pd.Series,
        *,
        layer: Optional[str] = None,
        aggregation: CytoTRACEAggregation.MEAN,
        ascending: bool = False,
        n_genes: int = 200,
    ) -> Tuple[np.ndarray, List[str]]:
        from sklearn.utils.sparsefuncs import csc_median_axis_0

        # fmt: off
        modifier = "negatively" if ascending else "positively"
        if n_genes <= 0:
            raise ValueError(f"Expected number of {modifier} correlated genes to be positive, found `{n_genes}`.")
        top_genes = [g for g in gene_corr.sort_values(ascending=ascending).index if g in self.adata.var_names]
        top_genes = top_genes[:n_genes]
        invalid_genes = int(gene_corr.loc[top_genes].isnull().sum())
        # fmt: on

        if invalid_genes:
            raise ValueError(
                f"Top `{len(top_genes)}` {modifier} correlated genes contain `{invalid_genes}` NaN values."
            )
        if not len(top_genes):
            raise ValueError("No genes have been selected.")

        if len(top_genes) != n_genes:
            logg.warning(
                f"Unable to get requested top {modifier} correlated `{n_genes}`. " f"Using top `{len(top_genes)}` genes"
            )

        # fmt: off
        imputed_exp = self.adata[:, top_genes].X if layer == "X" else self.adata[:, top_genes].layers[layer]
        if sp.issparse(imputed_exp) and aggregation not in (CytoTRACEAggregation.MEAN, CytoTRACEAggregation.MEDIAN):
            imputed_exp = imputed_exp.A

        if aggregation == CytoTRACEAggregation.MEAN:
            cytotrace_score = np.asarray(imputed_exp.mean(axis=1)).reshape((-1,))
        elif aggregation == CytoTRACEAggregation.MEDIAN:
            if sp.issparse(imputed_exp):
                cytotrace_score = np.asarray(csc_median_axis_0(imputed_exp.T.tocsc())).reshape((-1,))
            else:
                cytotrace_score = np.median(imputed_exp, axis=1)
        elif aggregation == CytoTRACEAggregation.GMEAN:
            cytotrace_score = st.gmean(imputed_exp, axis=1)
        elif aggregation == CytoTRACEAggregation.HMEAN:
            cytotrace_score = st.hmean(imputed_exp, axis=1)
        else:
            raise NotImplementedError(f"Aggregation method `{aggregation}` is not yet implemented.")
        # fmt: on
        return cytotrace_score, top_genes



import abc
from typing import Any, Optional

from anndata import AnnData

from cellrank._utils._docs import d

__all__ = ["AnnDataMixin"]


class AnnDataMixin(abc.ABC):
    """Mixin that allows for serialization from/to :class:`~anndata.AnnData`."""

    @abc.abstractmethod
    @d.dedent
    def _read_from_adata(self, adata: AnnData, **kwargs: Any) -> bool:
        """Populate attributes from :class:`~anndata.AnnData`.

        Parameters
        ----------
        %(adata)s
        kwargs
            Additional keyword arguments.

        Returns
        -------
        :obj:`True` if the de-serialization was successful, otherwise :obj:`False`.
        """
        return True

    @property
    @abc.abstractmethod
    def adata(self) -> AnnData:
        """Annotated data object."""

    @adata.setter
    @abc.abstractmethod
    def adata(self, adata: Optional[AnnData]) -> None:
        ...

    @abc.abstractmethod
    def __len__(self) -> int:
        ...

    @abc.abstractmethod
    @d.get_full_description(base="to_adata")
    def to_adata(self, **kwargs: Any) -> AnnData:
        """Serialize self to :class:`~anndata.Anndata`."""

    @classmethod
    @d.get_full_description(base="from_adata")
    @d.get_sections(base="from_adata", sections=["Returns"])
    @d.dedent
    def from_adata(cls, adata: AnnData, **kwargs: Any) -> "AnnDataMixin":
        """De-serialize self from :class:`~anndata.AnnData`.

        Parameters
        ----------
        %(adata)s
        kwargs
            Additional keyword arguments for initialization.

        Returns
        -------
        The de-serialized object.
        """
        obj = cls(adata, **kwargs)
        obj._read_from_adata(adata)

        return obj



from typing import Any, List, Optional, Sequence, Tuple, Union

import numpy as np
import pandas as pd
import scipy.stats as st
from pandas.api.types import infer_dtype

from matplotlib import cm, colors

from cellrank import logging as logg


def _create_colors(
    base_color: Union[str, Tuple[float, float, float]],
    n: int,
    hue_range: Optional[Tuple[float, float]] = (-0.1, 0.1),
    saturation_range: Optional[Tuple[float, float]] = (-0.3, 0.3),
    value_range: Optional[Tuple[float, float]] = (-0.3, 0.3),
    convert_to_rgb: bool = True,
    as_hex: bool = True,
) -> List[Any]:
    """Create variations of colors from base color.

    Parameters
    ----------
    base_color
        Base color which serves as a starting point.
    n
        Number of colors to create.
    hue_range
        Minimum and maximum value to add to the base color's hue.
        If `None`, don't adjust the hue.
    saturation_range
        Minimum and maximum value to add to the base color's saturation.
        If `None`, don't adjust the saturation.
    value_range
        Minimum and maximum value to add to the base color's value.
        If `None`, don't adjust the value.
    convert_to_rgb
        Whether to convert colors from HSV to RGB.
    as_hex:
        Whether to return colors as hex string.

    Returns
    -------
    List of colors, either as a hex string or an RGB array.
    """
    if not colors.is_color_like(base_color):
        raise ValueError("Base color is not color-like.")
    if n <= 0:
        raise ValueError(f"Number of colors must be > 0, found `{n}`.")

    base_color = colors.rgb_to_hsv(colors.to_rgb(base_color))

    if n == 1:
        res = [base_color]
    else:
        n *= 2  # sometimes the colors are too similar, we take every 2nd one
        res = np.repeat(base_color[..., np.newaxis], n, axis=1).T

        for i, r in enumerate((hue_range, saturation_range, value_range)):
            if r is None:
                continue
            r_low, r_high = sorted(r)
            c = base_color[i]

            res[:, i] = np.linspace(max(c + r_low, 0), min(c + r_high, 1), n)

    if convert_to_rgb:
        res = [colors.hsv_to_rgb(c) for c in res]
    if as_hex:
        res = [colors.to_hex(c) for c in res]

    return res[::2]  # we've created twice as many colors, select every other


def _convert_to_hex_colors(cols: Sequence[Any]) -> List[str]:
    if not all(colors.is_color_like(c) for c in cols):
        raise ValueError("Not all values are color-like.")

    return [colors.to_hex(c) for c in cols]


def _create_categorical_colors(n_categories: Optional[int] = None):
    from scanpy.plotting.palettes import vega_20_scanpy

    cmaps = [
        colors.ListedColormap(vega_20_scanpy),
        cm.Accent,
        colors.ListedColormap(np.array(cm.Dark2.colors)[[1, 2, 4, 5, 6]]),
        cm.Set1,
        cm.Set2,
        cm.Set3,
    ]
    max_cats = sum(c.N for c in cmaps)

    if n_categories is None:
        n_categories = max_cats
    if n_categories > max_cats:
        raise ValueError(f"Number of categories `{n_categories}` exceeded the maximum number of colors `{max_cats}`.")

    res = []
    for cmap in cmaps:
        res += [cmap(i) for i in range(cmap.N)][: n_categories - len(res)]
        if len(res) == n_categories:
            return _convert_to_hex_colors(res)

    raise RuntimeError(f"Unable to create `{n_categories}` colors.")


def _insert_categorical_colors(seen_colors: Union[np.ndarray, List], n_categories: int):
    seen_colors = set(_convert_to_hex_colors(seen_colors))
    candidates = list(filter(lambda c: c not in seen_colors, _create_categorical_colors()))[:n_categories]

    if len(candidates) != n_categories:
        raise RuntimeError(f"Unable to create `{n_categories}` categorical colors.")

    return candidates


def _contrasting_color(r: int, g: int, b: int) -> str:
    for val in [r, g, b]:
        assert 0 <= val <= 255

    return "#000000" if r * 0.299 + g * 0.587 + b * 0.114 > 186 else "#ffffff"


def _get_black_or_white(value: float, cmap) -> str:
    if np.isnan(value):
        return "#000000"
    if not (0.0 <= value <= 1.0):
        raise ValueError(f"Value must be in range `[0, 1]`, found `{value}`.")

    r, g, b, *_ = (int(c * 255) for c in cmap(value))
    return _contrasting_color(r, g, b)


def _get_bg_fg_colors(color, sat_scale: Optional[float] = None) -> Tuple[str, str]:
    if not colors.is_color_like(color):
        raise ValueError(f"Value `{color}` is not color-like.")

    color = np.squeeze(colors.to_rgba_array(color, alpha=1))[:3]
    if sat_scale is not None:
        h, s, v = colors.rgb_to_hsv(color)
        color = colors.hsv_to_rgb([h, s * sat_scale, v])

    return (
        colors.to_hex(color),
        _contrasting_color(*np.array(color * 255).astype(int)),
    )


def _map_names_and_colors(
    series_reference: pd.Series,
    series_query: pd.Series,
    colors_reference: Optional[np.array] = None,
    en_cutoff: Optional[float] = None,
) -> Union[pd.Series, Tuple[pd.Series, List[Any]]]:
    """Map annotations and colors from one series to another.

    Parameters
    ----------
    series_reference
        Series object with categorical annotations.
    series_query
        Series for which we would like to query the category names.
    colors_reference
        If given, colors for the query categories are pulled from this color array.
    en_cutoff
        In case of a non-perfect overlap between categories of the two series,
        this decides when to label a category in the query as 'Unknown'.

    Returns
    -------
    Series with updated category names and a corresponding array of colors.
    """
    # checks: dtypes, matching indices, make sure colors match the categories
    if not isinstance(series_reference.dtype, pd.CategoricalDtype):
        raise TypeError(f"Reference series must be `categorical`, found `{infer_dtype(series_reference)}`.")
    if not isinstance(series_query.dtype, pd.CategoricalDtype):
        raise TypeError(f"Query series must be `categorical`, found `{infer_dtype(series_query)}`.")
    if len(series_reference) != len(series_query):
        raise ValueError(
            f"Expected the reference and query to have the same length, "
            f"found `{len(series_reference)}`, `{len(series_query)}`."
        )

    if en_cutoff is not None and en_cutoff < 0:
        raise ValueError(f"Expected entropy cutoff to be non-negative, found `{en_cutoff}`.")

    if not np.all(series_reference.index == series_query.index):
        raise ValueError("Series indices do not match, cannot map names and colors.")

    process_colors = colors_reference is not None

    if not len(series_query):
        res = pd.Series([], dtype="category")
        return (res, []) if process_colors else res

    if process_colors:
        if len(colors_reference) < len(series_reference.cat.categories):
            raise ValueError(
                f"Length of reference colors `{len(colors_reference)}` is smaller than "
                f"length of reference categories `{len(series_reference.cat.categories)}`."
            )
        colors_reference = colors_reference[: len(series_reference.cat.categories)]
        if not all(colors.is_color_like(c) for c in colors_reference):
            raise ValueError("Not all values are valid colors.")
        if len(set(colors_reference)) != len(colors_reference):
            logg.warning("Color sequence contains non-unique elements")

    # create dataframe to store the associations between reference and query
    cats_query = series_query.cat.categories
    cats_reference = series_reference.cat.categories
    association_df = pd.DataFrame(None, index=cats_query, columns=cats_reference)

    # populate the dataframe - compute the overlap
    for cl in cats_query:
        row = [np.sum(series_reference.loc[np.array(series_query == cl)] == key) for key in cats_reference]
        association_df.loc[cl] = row
    association_df = association_df.apply(pd.to_numeric)

    # find the mapping which maximizes overlap
    names_query = association_df.T.idxmax()
    if en_cutoff is not None:
        association_df["entropy"] = st.entropy(association_df.T)
    association_df["name"] = names_query

    # assign query colors
    if process_colors:
        association_df["color"] = colors_query = [
            colors_reference[np.where(cats_reference == name)[0][0]] for name in names_query
        ]

    # next, we need to make sure that we have unique names and colors. In a first step, compute how many repetitions
    # we have
    names_query_series = pd.Series(names_query, dtype="category")
    frequ = {key: np.sum(names_query == key) for key in names_query_series.cat.categories}

    # warning: do NOT use np.array - if I pass for e.g. colors ['red'], the dtype will be '<U3'
    # but _create_colors convert them to hex, which will leave them trimmed to #ff or similar
    names_query_new = pd.Series(names_query.copy())
    if process_colors:
        colors_query_new = pd.Series(colors_query.copy())

    # Create unique names by adding suffixes "..._1, ..._2" etc and unique colors by shifting the original color
    for key, value in frequ.items():
        if value == 1:
            continue  # already unique, skip
        # deal with non-unique names
        unique_names = [f"{key}_{rep}" for rep in np.arange(1, value + 1)]
        # .value because of pandas 1.0.0
        names_query_new.iloc[(names_query_series == key).values] = unique_names
        if process_colors:
            color = association_df[association_df["name"] == key]["color"].values[0]
            shifted_colors = _create_colors(color, value, saturation_range=None)
            colors_query_new.iloc[(names_query_series == key).values] = shifted_colors

    # warnings: if it's categorical and assigning to `.cat.categories`, it will
    # take the categorical information, making the 2nd line below necessary
    names_query_new = names_query_new.astype("category")
    names_query_new = names_query_new.cat.reorder_categories(np.array(names_query_new))

    # issue a warning for mapping with high entropy
    if en_cutoff is not None:
        critical_cats = sorted(set(association_df.loc[association_df["entropy"] > en_cutoff, "name"].values))
        if len(critical_cats) > 0:
            logg.warning(f"The following states could not be mapped uniquely: `{critical_cats}`")

    return (names_query_new, list(_convert_to_hex_colors(colors_query_new))) if process_colors else names_query_new


def _compute_mean_color(cols: List[str]) -> str:
    """Compute mean color."""
    if not all(colors.is_color_like(c) for c in cols):
        raise ValueError(f"Not all values are valid colors `{cols}`.")

    cols = np.array([colors.rgb_to_hsv(colors.to_rgb(c)) for c in cols])

    return colors.to_hex(colors.hsv_to_rgb(np.mean(cols, axis=0)))



import abc
import enum
import functools
from typing import Any, Callable, Dict, Literal, Tuple, Type

__all__ = ["ModeEnum", "DEFAULT_BACKEND"]

DEFAULT_BACKEND = "loky"
Backend_t = Literal["loky", "multiprocessing", "threading"]


class PrettyEnum(enum.Enum):
    """Enum with a pretty :meth:`__str__` and :meth:`__repr__`."""

    @property
    def v(self) -> Any:
        """Alias for :attr`value`."""
        return self.value

    def __repr__(self) -> str:
        return f"{self.value!r}"

    def __str__(self) -> str:
        return f"{self.value!s}"


def _pretty_raise_enum(cls: Type["ErrorFormatterABC"], func: Callable) -> Callable:
    @functools.wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> "ErrorFormatterABC":
        try:
            return func(*args, **kwargs)
        except ValueError as e:
            _cls, value, *_ = args
            e.args = (cls._format(value),)
            raise e

    if not issubclass(cls, ErrorFormatterABC):
        raise TypeError(f"Class `{cls}` must be subtype of `ErrorFormatterABC`.")
    if not len(cls.__members__):
        # empty enum, for class hierarchy
        return func

    return wrapper


class ABCEnumMeta(enum.EnumMeta, abc.ABCMeta):  # noqa: D101
    def __call__(cls, *args, **kwargs):  # noqa
        if getattr(cls, "__error_format__", None) is None:
            raise TypeError(f"Can't instantiate class `{cls.__name__}` " f"without `__error_format__` class attribute.")
        return super().__call__(*args, **kwargs)

    def __new__(cls, clsname: str, superclasses: Tuple[type], attributedict: Dict[str, Any]):  # noqa: D102
        res = super().__new__(cls, clsname, superclasses, attributedict)
        res.__new__ = _pretty_raise_enum(res, res.__new__)
        return res


class ErrorFormatterABC(abc.ABC):  # noqa: D101
    __error_format__ = "Invalid option `{!r}` for `{}`. Valid options are: `{}`."

    @classmethod
    def _format(cls, value) -> str:
        return cls.__error_format__.format(value, cls.__name__, [m.value for m in cls.__members__.values()])


class ModeEnum(str, ErrorFormatterABC, PrettyEnum, metaclass=ABCEnumMeta):  # noqa: D101
    def _generate_next_value_(self, start, count, last_values):
        return str(self).lower()



import warnings
from typing import Any, Optional

import scvelo as scv
from scvelo.tools.velocity_embedding import quiver_autoscale

import numpy as np
import scipy.sparse as sp

from cellrank import logging as logg
from cellrank._utils._docs import d
from cellrank._utils._key import Key
from cellrank.kernels._utils import _get_basis

__all__ = ["TmatProjection"]


class TmatProjection:
    """Project transition matrix onto a low-dimensional embedding.

    Should be used for visualization purposes.

    Parameters
    ----------
    kexpr
        Kernel that contains a transition matrix.
    basis
        Key in :attr:`~anndata.AnnData.obsm` where the basis is stored.
    """

    def __init__(self, kexpr: "ExpernelExpression", basis: str = "umap"):  # noqa: F821
        from cellrank.kernels.mixins import ConnectivityMixin

        for kernel in kexpr.kernels:
            if not isinstance(kernel, ConnectivityMixin):
                logg.warning(
                    f"{kernel!r} is not a kNN based kernel. "
                    f"The embedding projection works best for kNN-based kernels"
                )
                break
        self._kexpr = kexpr
        self._basis = basis[2:] if basis.startswith("X_") else basis
        self._key: Optional[str] = None

    def project(
        self,
        key_added: Optional[str] = None,
        recompute: bool = False,
        connectivities: Optional[sp.spmatrix] = None,
    ) -> None:
        """Project transition matrix onto an embedding.

        This function is adapted from :func:`~scvelo.tl.velocity_embedding`.

        Parameters
        ----------
        key_added
            Key in :attr:`~anndata.AnnData.obsm` where to store the projection.
        recompute
            Whether to recompute an existing projection.
        connectivities
            Connectivity matrix to use for projection. If :obj:`None`, use ones from the underlying kernel, is possible.

        Returns
        -------
        Nothing, just updates :attr:`adata` with the projection and the parameters used for computation.
        """
        from cellrank.kernels.mixins import ConnectivityMixin

        self._key = Key.uns.kernel(self._kexpr.backward, key=key_added)
        ukey = f"{self._key}_params"
        key = f"{self._key}_{self._basis}"

        if not recompute and key in self._kexpr.adata.obsm:
            logg.info(f"Using precomputed projection `adata.obsm[{key!r}]`")
            return

        start = logg.info(f"Projecting transition matrix onto `{self._basis}`")
        if connectivities is None:
            try:
                connectivities, *_ = (c.connectivities for c in self._kexpr.kernels if isinstance(c, ConnectivityMixin))
            except ValueError:
                raise RuntimeError(
                    "Unable to find connectivities in the kernel. "
                    "Please supply them explicitly as `connectivities=...`."
                ) from None
        if not sp.isspmatrix_csr(connectivities):
            connectivities = connectivities.tocsr()

        emb = _get_basis(self._kexpr.adata, self._basis)
        T_emb = np.empty_like(emb)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            for row_id, row in enumerate(self._kexpr.transition_matrix):
                conn_idxs = connectivities[row_id, :].indices
                dX = emb[conn_idxs] - emb[row_id, None]

                if np.any(np.isnan(dX)):
                    T_emb[row_id, :] = np.nan
                else:
                    probs = row[:, conn_idxs].A.squeeze() if sp.issparse(row) else row[conn_idxs]
                    dX /= np.linalg.norm(dX, axis=1)[:, None]
                    dX = np.nan_to_num(dX)
                    T_emb[row_id, :] = probs.dot(dX) - dX.sum(0) / dX.shape[0]

        T_emb /= 3 * quiver_autoscale(np.nan_to_num(emb), T_emb)

        embs = self._kexpr.adata.uns.get(ukey, {}).get("embeddings", [])
        if self._basis not in embs:
            embs = list(embs) + [self._basis]
            self._kexpr.adata.uns[ukey] = self._kexpr.adata.uns.get(ukey, {})
            self._kexpr.adata.uns[ukey]["embeddings"] = embs

        logg.info(
            f"Adding `adata.obsm[{key!r}]`\n    Finish",
            time=start,
        )
        self._kexpr.adata.obsm[key] = T_emb

    @d.dedent
    def plot(self, *args: Any, stream: bool = True, **kwargs: Any) -> None:
        """Plot projected transition matrix in a embedding.

        Parameters
        ----------
        args
            Positional argument for the plotting function.
        stream
            If :obj:`True``, use :func:`~scvelo.pl.velocity_embedding_stream`.
            Otherwise, use :func:`~scvelo.pl.velocity_embedding_grid`.
        kwargs
            Keyword argument for the above-mentioned plotting function.

        Returns
        -------
        %(just_plots)s
        """
        if stream:
            return scv.pl.velocity_embedding_stream(
                self._kexpr.adata, *args, basis=self._basis, vkey=self._key, **kwargs
            )
        return scv.pl.velocity_embedding_grid(self._kexpr.adata, *args, basis=self._basis, vkey=self._key, **kwargs)



import itertools
import pathlib
from typing import Any, List, Literal, Mapping, Optional, Sequence, Tuple, Union

import scvelo as scv

import numpy as np
import pandas as pd
import scipy.sparse as sp
from pandas.api.types import infer_dtype

import matplotlib.pyplot as plt
from matplotlib.collections import LineCollection
from matplotlib.colors import LinearSegmentedColormap, to_hex

from anndata import AnnData

from cellrank import logging as logg
from cellrank._utils._docs import d
from cellrank._utils._parallelize import parallelize
from cellrank._utils._utils import save_fig
from cellrank.kernels._utils import _get_basis

__all__ = ["RandomWalk"]

Indices_t = Optional[Union[Sequence[str], Mapping[str, Union[str, Sequence[str], Tuple[float, float]]]]]


@d.dedent
class RandomWalk:
    """Class that simulates a random walk on a Markov chain.

    Parameters
    ----------
    %(adata)s
    transition_matrix
        Row-stochastic transition matrix.
    start_ixs
        Indices from which to uniformly sample the starting points. If :obj:`None`, use all points.
    stop_ixs
        Indices which when hit, the random walk is terminated.
    """

    def __init__(
        self,
        adata: AnnData,
        transition_matrix: Union[np.ndarray, sp.spmatrix],
        start_ixs: Optional[Sequence[int]] = None,
        stop_ixs: Optional[Sequence[int]] = None,
    ):
        if transition_matrix.ndim != 2 or (transition_matrix.shape[0] != transition_matrix.shape[1]):
            raise ValueError(f"Expected transition matrix to be a square matrix, found `{transition_matrix.ndim}`.")
        if transition_matrix.shape[0] != adata.n_obs:
            raise ValueError(
                f"Expected transition matrix to be of shape `{adata.n_obs, adata.n_obs}`,"
                f"found `{transition_matrix.shape}`."
            )
        if not np.allclose(transition_matrix.sum(1), 1.0):
            raise ValueError("Transition matrix is not row-stochastic.")

        self._adata = adata
        self._tmat = transition_matrix
        self._ixs = np.arange(self._tmat.shape[0])
        self._is_sparse = sp.issparse(self._tmat)

        start_ixs = self._normalize_ixs(start_ixs, kind="start")
        stop_ixs = self._normalize_ixs(stop_ixs, kind="stop")
        self._stop_ixs = set([] if stop_ixs is None or not len(stop_ixs) else stop_ixs)
        self._starting_dist = np.ones_like(self._ixs) if start_ixs is None else np.isin(self._ixs, start_ixs)
        _sum = np.sum(self._starting_dist)
        if _sum == 0:
            raise ValueError("No starting indices have been selected.")

        self._starting_dist = self._starting_dist.astype(transition_matrix.dtype) / _sum

    @d.get_sections(base="rw_sim", sections=["Parameters"])
    def simulate_one(
        self,
        max_iter: Union[int, float] = 0.25,
        seed: Optional[int] = None,
        successive_hits: int = 0,
    ) -> np.ndarray:
        """Simulate one random walk.

        Parameters
        ----------
        max_iter
            Maximum number of steps of a random walk. If a :class:`float`, it can be specified
            as a fraction of the number of cells.
        seed
            Random seed.
        successive_hits
            Number of successive hits in the ``stop_ixs`` required to stop prematurely.

        Returns
        -------
        Array of shape ``(max_iter + 1,)`` of states that have been visited.
        If ``stop_ixs`` was specified, the array may have a smaller shape.
        """
        max_iter = self._max_iter(max_iter)
        if successive_hits < 0:
            raise ValueError(f"Expected number of successive hits to be positive, found `{successive_hits}`.")

        rs = np.random.RandomState(seed)
        ix = rs.choice(self._ixs, p=self._starting_dist)
        sim, cnt = [ix], -1

        for _ in range(max_iter):
            ix = self._sample(ix, rs=rs)
            sim.append(ix)
            cnt = (cnt + 1) if self._should_stop(ix) else -1
            if cnt >= successive_hits:
                break

        return np.array(sim)

    def _simulate_many(
        self,
        sims: np.ndarray,
        max_iter: Union[int, float] = 0.25,
        seed: Optional[int] = None,
        successive_hits: int = 0,
        queue: Optional[Any] = None,
    ) -> List[np.ndarray]:
        res = []
        for s in sims:
            sim = self.simulate_one(
                max_iter=max_iter,
                seed=None if seed is None else seed + s,
                successive_hits=successive_hits,
            )
            res.append(sim)
            if queue is not None:
                queue.put(1)

        if queue is not None:
            queue.put(None)

        return res

    @d.dedent
    def simulate_many(
        self,
        n_sims: int,
        max_iter: Union[int, float] = 0.25,
        seed: Optional[int] = None,
        successive_hits: int = 0,
        n_jobs: Optional[int] = None,
        backend: str = "loky",
        show_progress_bar: bool = True,
    ) -> List[np.ndarray]:
        """Simulate many random walks.

        Parameters
        ----------
        n_sims
            Number of random walks to simulate.
        %(rw_sim.params)s
        %(parallel)s

        Returns
        -------
        List of arrays of shape ``(max_iter + 1,)`` of states that have been visited.
        If ``stop_ixs`` was specified, the arrays may have smaller shape.
        """
        if n_sims <= 0:
            raise ValueError(f"Expected number of simulations to be positive, found `{n_sims}`.")
        max_iter = self._max_iter(max_iter)
        start = logg.info(f"Simulating `{n_sims}` random walks of maximum length `{max_iter}`")

        simss = parallelize(
            self._simulate_many,
            collection=np.arange(n_sims),
            n_jobs=n_jobs,
            backend=backend,
            show_progress_bar=show_progress_bar,
            as_array=False,
            unit="sim",
        )(max_iter=max_iter, seed=seed, successive_hits=successive_hits)
        simss = list(itertools.chain.from_iterable(simss))

        logg.info("    Finish", time=start)

        return simss

    @d.dedent
    def plot(
        self,
        sims: List[np.ndarray],
        basis: str = "umap",
        cmap: Union[str, LinearSegmentedColormap] = "gnuplot",
        linewidth: float = 1.0,
        linealpha: float = 0.3,
        ixs_legend_loc: Optional[str] = None,
        figsize: Optional[Tuple[float, float]] = None,
        dpi: Optional[int] = None,
        save: Optional[Union[str, pathlib.Path]] = None,
        **kwargs: Any,
    ) -> None:
        """Plot simulated random walks.

        Parameters
        ----------
        sims
            The simulated random walks.
        basis
            Basis used for plotting.
        cmap
            Colormap for the random walks.
        linewidth
            Line width for the random walks.
        linealpha
            Line alpha.
        ixs_legend_loc
            Position of the legend describing start- and endpoints.
        %(plotting)s
        kwargs
            Keyword arguments for :func:`~scvelo.pl.scatter`.

        Returns
        -------
        %(just_plots)s
        """
        emb = _get_basis(self._adata, basis)
        if isinstance(cmap, str):
            cmap = plt.get_cmap(cmap)
        if not isinstance(cmap, LinearSegmentedColormap):
            if not hasattr(cmap, "colors"):
                raise AttributeError("Unable to create a colormap, `cmap` does not have attribute `colors`.")
            cmap = LinearSegmentedColormap.from_list(
                "random_walk",
                colors=cmap.colors,
                N=max(map(len, sims)),
            )

        fig, ax = plt.subplots(figsize=figsize, dpi=dpi)
        scv.pl.scatter(self._adata, basis=basis, show=False, ax=ax, **kwargs)

        logg.info("Plotting random walks")
        for sim in sims:
            x = emb[sim][:, 0]
            y = emb[sim][:, 1]
            points = np.array([x, y]).T.reshape(-1, 1, 2)
            segments = np.concatenate([points[:-1], points[1:]], axis=1)
            n_seg = len(segments)

            lc = LineCollection(
                segments,
                linewidths=linewidth,
                colors=[cmap(float(i) / n_seg) for i in range(n_seg)],
                alpha=linealpha,
                zorder=2,
            )
            ax.add_collection(lc)

        for ix in [0, -1]:
            ixs = [sim[ix] for sim in sims]
            from scvelo.plotting.utils import default_size, plot_outline

            plot_outline(
                x=emb[ixs][:, 0],
                y=emb[ixs][:, 1],
                outline_color=("black", to_hex(cmap(float(abs(ix))))),
                kwargs={
                    "s": kwargs.get("size", default_size(self._adata)) * 1.1,
                    "alpha": 0.9,
                },
                ax=ax,
                zorder=4,
            )

        if ixs_legend_loc not in (None, "none"):
            from cellrank.pl._utils import _position_legend

            h1 = ax.scatter([], [], color=cmap(0.0), label="start")
            h2 = ax.scatter([], [], color=cmap(1.0), label="stop")
            legend = ax.get_legend()
            if legend is not None:
                ax.add_artist(legend)
            _position_legend(ax, legend_loc=ixs_legend_loc, handles=[h1, h2])

        if save is not None:
            save_fig(fig, save)

    def _normalize_ixs(self, ixs: Indices_t, *, kind: Literal["start", "stop"]) -> Optional[np.ndarray]:
        if ixs is None:
            return None

        if isinstance(ixs, dict):
            # fmt: off
            if len(ixs) != 1:
                raise ValueError(f"Expected to find only 1 cluster key, found `{len(ixs)}`.")
            key = next(iter(ixs.keys()))
            if key not in self._adata.obs:
                raise KeyError(f"Unable to find data in `adata.obs[{key!r}]`.")

            vals = self._adata.obs[key]
            if isinstance(vals.dtype, pd.CategoricalDtype):
                ixs = np.where(np.isin(vals, ixs[key]))[0]
            elif np.issubdtype(vals.dtype, np.number):
                if len(ixs[key]) != 2:
                    raise ValueError(f"Expected range to be of length `2`, found `{len(ixs[key])}`")
                minn, maxx = sorted(ixs[key])
                ixs = np.where((vals >= minn) & (vals <= maxx))[0]
            else:
                raise TypeError(f"Expected `adata.obs[{key!r}]` to be numeric or categorical, "
                                f"found `{infer_dtype(vals)}`.")
            # fmt: on
        elif isinstance(ixs, str):
            ixs = np.where(self._adata.obs_names == ixs)[0]
        elif isinstance(ixs[0], str):
            ixs = np.where(np.isin(self._adata.obs_names, ixs))[0]
        elif isinstance(ixs[0], bool):
            if len(ixs) != self._adata.n_obs:
                raise ValueError(
                    f"Expected `bool` {kind} indices of length" f"`{self._adata.n_obs}`, found `{len(ixs)}`."
                )
            ixs = np.where(ixs)[0]
        elif isinstance(ixs[0], int):
            ixs = list(set(ixs))
            if max(ixs) >= self._adata.n_obs:
                raise IndexError(max(ixs))
            if min(ixs) < -self._adata.n_obs:
                raise IndexError(min(ixs))
        else:
            raise TypeError(
                f"Expected {kind} indices to be either `dict` or a sequence of "
                f"`int`, `str`, `bool`, found `{type(ixs).__nam__}`."
            )

        if not len(ixs):
            raise ValueError(f"No {kind} indices have been selected.")

        return ixs

    def _should_stop(self, ix: int) -> bool:
        return ix in self._stop_ixs

    def _sample(self, ix: int, *, rs: np.random.RandomState) -> int:
        return rs.choice(
            self._ixs,
            p=self._tmat[ix].A.squeeze() if self._is_sparse else self._tmat[ix],
        )

    def _max_iter(self, max_iter: Union[int, float]) -> int:
        if isinstance(max_iter, float):
            max_iter = int(np.ceil(max_iter * len(self._ixs)))
        if max_iter <= 1:
            raise ValueError(f"Expected number of iterations to be > 1, found `{max_iter}`.")
        return max_iter



from cellrank.logging._logging import (
    debug,
    error,
    hint,
    info,
    print_version_and_date,
    print_versions,
    warning,
)

__all__ = [
    "debug",
    "error",
    "hint",
    "info",
    "print_version_and_date",
    "print_versions",
    "warning",
]



import copy
import logging

from scanpy import settings

from cellrank.logging._logging import _LogFormatter, _RootLogger

__all__ = ["settings"]


def _set_log_file(settings):
    file = settings.logfile
    name = settings.logpath
    root = settings._root_logger
    h = logging.StreamHandler(file) if name is None else logging.FileHandler(name)
    h.setFormatter(_LogFormatter())
    h.setLevel(root.level)

    if len(root.handlers) == 1:
        root.removeHandler(root.handlers[0])
    elif len(root.handlers) > 1:
        raise RuntimeError("CellRank's root logger somehow got more than one handler.")

    root.addHandler(h)


settings = copy.copy(settings)
settings._root_logger = _RootLogger(settings.verbosity)
# these 2 lines are necessary to get it working (otherwise no logger is found)
# this is a hacky way of modifying the logging, in the future, use our own
_set_log_file(settings)
settings.verbosity = settings.verbosity
